"""Wherein functionality concerning listing basket jsons is contained."""

def _get_list_of_basket_jsons(root_dir, file_system):
    return [x for x in file_system.find(root_dir)
            if x.endswith("basket_manifest.json")]



"""Contains scripts concerning the Basket class.
"""

import json
import os
from pathlib import Path

import pandas as pd

from .config import get_file_system, prohibited_filenames


class BasketInitializer:
    """Initializes basket class. Validates input args.
    """

    def __init__(self, basket_address, **kwargs):
        """Handles set up of basket. Calls validation.

        Parameters
        ----------
        basket_address: str
            Argument can take one of two forms: either a path to the basket
            directory, or the UUID of the basket.
        **file_system: fsspec object (optional)
            The fsspec filesystem to be used for retrieving and uploading. This
            is only used when basket_address is a path.
        **pantry: weave.Pantry (optional)
            The pantry which the basket uuid is associated with. Only for UUID.
        """

        if "pantry" in kwargs:
            self.file_system = kwargs["pantry"].file_system
        else:
            self.file_system = kwargs.get("file_system", get_file_system())
        try:
            self._set_up_basket_from_path(basket_address)
        except ValueError as error:
            if str(error) != f"Basket does not exist: {self.basket_path}":
                raise error

            if "pantry" not in kwargs:
                raise KeyError("pantry, required to set up basket from UUID,"
                               "is not in kwargs.") from error
            self.set_up_basket_from_uuid(basket_address, kwargs["pantry"])

        self.manifest_path = f"{self.basket_path}/basket_manifest.json"
        self.supplement_path = f"{self.basket_path}/basket_supplement.json"
        self.metadata_path = f"{self.basket_path}/basket_metadata.json"
        self.validate()

    def _set_up_basket_from_path(self, basket_address):
        """Attempts to set up a basket from a filepath.

        Paramters
        ---------
        basket_address: str
            Argument can take one of two forms: either a path to the basket
            directory or the UUID of the basket. In this case it is assumed to
            be a path to the basket directory.
        """

        self.basket_path = os.fspath(basket_address)
        self.validate_basket_path()

    def set_up_basket_from_uuid(self, basket_address, pantry):
        """Attempts to set up a basket from a uuid.

        Note that if the basket cannot be set up from a uuid then an attempt to
        set up the basket from a filepath will be made.

        Parameters
        ----------
        basket_address: str
            Argument can take one of two forms: either a path to the basket
            directory or the UUID of the basket. In this case it is assumed to
            be the UUID of the basket.
        pantry: weave.Pantry
            The pantry which the basket uuid is associated with.
        """

        try:
            row = pantry.index.get_rows(basket_address)
            pantry.validate_path_in_pantry(row.iloc[0].address)
            self._set_up_basket_from_path(basket_address=row.iloc[0].address)
        except BaseException as error:
            self.basket_path = basket_address

            pantry_error_msg = "Attempting to access basket outside of pantry:"
            if str(error).startswith(pantry_error_msg):
                raise error
            self.validate_basket_path()
            # The above line should raise an exception
            # The below line is more or less a fail safe and will raise the
            # exception
            raise error

    def validate_basket_path(self):
        """Validates basket exists."""
        if not self.file_system.exists(self.basket_path):
            raise ValueError(f"Basket does not exist: {self.basket_path}")

    def validate(self):
        """Validates basket health."""
        if not self.file_system.exists(self.manifest_path):
            raise FileNotFoundError(
                f"Invalid Basket, basket_manifest.json "
                f"does not exist: {self.manifest_path}"
            )

        if not self.file_system.exists(self.supplement_path):
            raise FileNotFoundError(
                f"Invalid Basket, basket_supplement.json "
                f"does not exist: {self.supplement_path}"
            )

# Ignoring because there is a necessary and
# reasonable amount of variables in this case
# pylint: disable=too-many-instance-attributes
class Basket(BasketInitializer):
    """This class provides convenience functions for accessing basket contents.
    """

    def __init__(self, basket_address, **kwargs):
        """Initializes the Basket_Class.

        If basket_address is a path, the basket will be loaded directly using
        the file_system or pantry. If basket_address is a UUID, the basket will
        be loaded using the provided pantry's index.

        Parameters
        ----------
        basket_address: str
            Argument can take one of two forms: either a path to the Basket
            directory or the UUID of the basket.
        **file_system: fsspec object (optional)
            The fsspec filesystem to be used for retrieving and uploading. May
            be used if basket_address is a path.
        **pantry: weave.Pantry (optional)
            The pantry which the basket uuid is associated with. Required if
            basket_address is a UUID.
        """

        super().__init__(basket_address, **kwargs)

        self.manifest = None
        self.supplement = None
        self.metadata = None
        self.get_manifest()
        self.populate_members()

    def populate_members(self):
        """Populate the Basket_Class member variables."""
        self.uuid = self.manifest["uuid"]
        self.upload_time = self.manifest["upload_time"]
        self.parent_uuids = self.manifest["parent_uuids"]
        self.basket_type = self.manifest["basket_type"]
        self.label = self.manifest["label"]
        self.weave_version = self.manifest.get("weave_version", "<0.13.0")
        self.address = self.basket_path
        self.storage_type = self.file_system.__class__.__name__

    def get_manifest(self):
        """Return basket_manifest.json as a python dictionary."""
        if self.manifest is not None:
            return self.manifest

        with self.file_system.open(self.manifest_path, "rb") as file:
            self.manifest = json.load(file)
            return self.manifest

    def get_supplement(self):
        """Return basket_supplement.json as a python dictionary."""
        if self.supplement is not None:
            return self.supplement

        with self.file_system.open(self.supplement_path, "rb") as file:
            self.supplement = json.load(file)
            return self.supplement

    def get_metadata(self):
        """Return basket_metadata.json as a python dictionary.

        Return None if metadata doesn't exist.
        """
        if self.metadata is not None:
            return self.metadata

        if self.file_system.exists(self.metadata_path):
            with self.file_system.open(self.metadata_path, "rb") as file:
                self.metadata = json.load(file)
                return self.metadata
        else:
            return None

    # Disabling pylint name warning for ls, as it is the standard name
    # for functions of it's type in the computing world. It makes
    # sense to continue to name this function ls.
    # pylint: disable-next=invalid-name
    def ls(self, relative_path=None):
        """List directories and files in the basket.

           Call filesystem.ls relative to the basket directory.
           When relative_path = None, filesystem.ls is invoked
           from the base directory of the basket. If there are folders
           within the basket, relative path can be used to observe contents
           within folders.

           Example: if there exists a folder with the
           name 'folder1' within the basket, 'folder1' can be passed
           as the relative path to get back the filesystem.ls results
           of 'folder1'.

        Parameters
        ----------
        relative_path: str (default=None)
            relative path in the basket to pass to filesystem.ls.

        Returns
        ---------
        filesystem.ls results of the basket.
        """

        ls_path = os.fspath(Path(self.basket_path))

        if relative_path is not None:
            relative_path = os.fspath(relative_path)
            ls_path = os.fspath(
              Path(os.path.join(self.basket_path, relative_path))
            )

        if ls_path == os.fspath(Path(self.basket_path)):
            # Remove any prohibited files from the list if they exist
            # in the root directory
            # Note that file_system.ls can have unpredictable behavior if
            # not passing refresh=True
            ls_results = self.file_system.ls(ls_path, refresh=True)
            return [
                x
                for x in ls_results
                if os.path.basename(Path(x)) not in prohibited_filenames
            ]
        # Note that file_system.ls can have unpredictable behavior if
        # not passing refresh=True
        return self.file_system.ls(ls_path, refresh=True)

    def to_pandas_df(self):
        """Return a dataframe of the basket member variables."""
        data = [self.uuid, self.upload_time,
                self.parent_uuids, self.basket_type,
                self.label, self.weave_version, self.address,
                self.storage_type]
        columns = ["uuid", "upload_time", "parent_uuids",
                   "basket_type", "label", "weave_version",
                   "address", "storage_type"]

        return pd.DataFrame(data=[data], columns=columns)



"""Pytest tests for basket.py related functionality."""

import json
import os
import re
from pathlib import Path

import pytest
import s3fs
import pandas as pd
from fsspec.implementations.local import LocalFileSystem

from weave.__init__ import __version__ as weave_version
from weave.basket import Basket
from weave.index.create_index import create_index_from_fs
from weave.pantry import Pantry
from weave.index.index_pandas import IndexPandas
from weave.tests.pytest_resources import PantryForTest

###############################################################################
#                      Pytest Fixtures Documentation:                         #
#            https://docs.pytest.org/en/7.3.x/how-to/fixtures.html            #
#                                                                             #
#                  https://docs.pytest.org/en/7.3.x/how-to/                   #
#          fixtures.html#teardown-cleanup-aka-fixture-finalization            #
#                                                                             #
#  https://docs.pytest.org/en/7.3.x/how-to/fixtures.html#fixture-parametrize  #
###############################################################################

# Pylint doesn't like redefining the test fixture here from
# test_basket, but this is the right way to do this if at some
# point in the future the two need to be differentiated.
# pylint: disable=duplicate-code

s3fs = s3fs.S3FileSystem(
    client_kwargs={"endpoint_url": os.environ["S3_ENDPOINT"]}
)
local_fs = LocalFileSystem()


# Test with two different fsspec file systems (above).
@pytest.fixture(
    name="test_pantry",
    params=[s3fs, local_fs],
    ids=["S3FileSystem", "LocalFileSystem"],
)
def fixture_test_pantry(request, tmpdir):
    """Fixture to set up and tear down test_basket."""
    file_system = request.param
    test_pantry = PantryForTest(tmpdir, file_system)
    yield test_pantry
    test_pantry.cleanup_pantry()


def test_basket_basket_path_is_pathlike():
    """Test that an error is returned when trying to instantiate a basket with
    invalid basket address type.
    """

    basket_path = 1
    with pytest.raises(
        TypeError,
        match="expected str, bytes or os.PathLike object, not int",
    ):
        Basket(basket_path)


def test_basket_address_does_not_exist(test_pantry):
    """Test that an error is raised when trying to instantiate a basket with an
    invalid basket address.
    """

    basket_path = Path("i n v a l i d p a t h")
    pantry = Pantry(IndexPandas,
                    pantry_path=test_pantry.pantry_path,
                    file_system=test_pantry.file_system)
    pantry.index.generate_index()
    with pytest.raises(
        ValueError, match=f"Basket does not exist: {basket_path}"
    ):
        Basket(
            Path(basket_path),
            file_system=test_pantry.file_system,
            pantry=pantry
        )


def test_make_basket_with_uuid_stays_in_pantry(test_pantry):
    """Tests the pantry does not access baskets outside of itself."""
    tmp_basket_dir_one = test_pantry.set_up_basket("basket_one")
    test_pantry.upload_basket(tmp_basket_dir=tmp_basket_dir_one, uid="0001")

    pantry = Pantry(
        IndexPandas,
        pantry_path=test_pantry.pantry_path,
        file_system=test_pantry.file_system
    )

    # Save and delete the basket.
    pantry.index.generate_index()
    index = pantry.index.to_pandas_df()
    pantry.index.untrack_basket(index.iloc[0].uuid)

    # Modify the basket address to a new (fake) pantry.
    address = index.iloc[0].address
    address = address.split(os.path.sep)
    address[0] += "-2"
    new_address = (os.path.sep).join(address)
    index.at[0,"address"] = new_address

    # Track the new basket
    pantry.index.track_basket(index)

    error_msg = f"Attempting to access basket outside of pantry: {new_address}"
    with pytest.raises(ValueError, match=error_msg):
        Basket(index.iloc[0].uuid, pantry=pantry)


def test_basket_no_manifest_file(test_pantry):
    """Test that an error is raised when attempting to instantiate a basket
    with a missing basket manifest file.
    """

    # Create a temporary basket with a test file, and upload it.
    tmp_basket_dir_name = "test_basket_tmp_dir"
    tmp_basket_dir = test_pantry.set_up_basket(tmp_basket_dir_name)
    basket_path = test_pantry.upload_basket(tmp_basket_dir=tmp_basket_dir)

    # Manually remove the basket_manifest file.
    manifest_path = os.path.join(basket_path, "basket_manifest.json")
    test_pantry.file_system.rm(manifest_path)

    # Attempt to create a Basket from the malformed basket (missing manifest)
    with pytest.raises(
        FileNotFoundError,
        match=(
            "Invalid Basket, basket_manifest.json "
            + f"does not exist: {manifest_path}"
        ),
    ):
        Basket(
            Path(basket_path),
            file_system=test_pantry.file_system
        )


def test_basket_no_supplement_file(test_pantry):
    """Test that an error is raised when attempting to instantiate a basket
    with a missing basket supplement file.
    """

    # Create a temporary basket with a test file, and upload it.
    tmp_basket_dir_name = "test_basket_tmp_dir"
    tmp_basket_dir = test_pantry.set_up_basket(tmp_basket_dir_name)
    basket_path = test_pantry.upload_basket(tmp_basket_dir=tmp_basket_dir)

    # Manually remove the basket_supplement file.
    supplement_path = os.path.join(basket_path, "basket_supplement.json")
    test_pantry.file_system.rm(supplement_path)

    # Attempt to create a Basket from the malformed basket (missing supplement)
    with pytest.raises(
        FileNotFoundError,
        match=(
            "Invalid Basket, basket_supplement.json "
            + f"does not exist: {supplement_path}"
        ),
    ):
        Basket(
            Path(basket_path),
            file_system=test_pantry.file_system
        )


def test_basket_get_manifest(test_pantry):
    """Test that the manifest of an uploaded basket is correctly retrieved
    using the get_manifest function.
    """

    # Create a temporary basket with a test file, and upload it.
    tmp_basket_dir_name = "test_basket_tmp_dir"
    tmp_basket_dir = test_pantry.set_up_basket(tmp_basket_dir_name)
    basket_path = test_pantry.upload_basket(tmp_basket_dir=tmp_basket_dir)

    basket = Basket(
        Path(basket_path),
        file_system=test_pantry.file_system
    )
    manifest = basket.get_manifest()
    assert manifest == {
        "uuid": "0000",
        "parent_uuids": [],
        "basket_type": "test_basket",
        "label": "",
        "upload_time": manifest["upload_time"],
        "weave_version": weave_version,
    }


def test_basket_get_manifest_cached(test_pantry):
    """Test that the get_manifest function retreives the cached copy."""
    # Create a temporary basket with a test file, and upload it.
    tmp_basket_dir_name = "test_basket_tmp_dir"
    tmp_basket_dir = test_pantry.set_up_basket(tmp_basket_dir_name)
    basket_path = test_pantry.upload_basket(tmp_basket_dir=tmp_basket_dir)

    basket = Basket(
        Path(basket_path),
        file_system=test_pantry.file_system
    )

    # Read the basket_manifest.json file and store as a dictionary for later.
    manifest = basket.get_manifest()
    manifest_path = basket.manifest_path

    # Manually replace the manifest file.
    test_pantry.file_system.rm(manifest_path)
    with test_pantry.file_system.open(manifest_path, "w") as outfile:
        json.dump({"junk": "b"}, outfile)

    # Manifest should already be stored and the new file shouldn't be read.
    manifest = basket.get_manifest()
    assert manifest == {
        "uuid": "0000",
        "parent_uuids": [],
        "basket_type": "test_basket",
        "label": "",
        "upload_time": manifest["upload_time"],
        "weave_version": weave_version,
    }


def test_basket_get_supplement(test_pantry):
    """Test that the get_supplement function returns the expected values."""
    # Create a temporary basket with a test file, and upload it.
    tmp_basket_dir_name = "test_basket_tmp_dir"
    tmp_basket_dir = test_pantry.set_up_basket(tmp_basket_dir_name)
    basket_path = test_pantry.upload_basket(tmp_basket_dir=tmp_basket_dir)

    basket = Basket(
        Path(basket_path),
        file_system=test_pantry.file_system
    )

    # Create a copy of the basket's expected upload_items.
    upload_items = [{"path": str(tmp_basket_dir.realpath()), "stub": False}]

    # Check the expected values are the same as the actual values.
    supplement = basket.get_supplement()
    assert supplement == {
        "upload_items": upload_items,
        "integrity_data": supplement["integrity_data"],
    }


def test_basket_get_supplement_cached(test_pantry):
    """Test that the get_supplement function retrieves cached copies of a
    basket's supplement.
    """

    # Create a temporary basket with a test file, and upload it.
    tmp_basket_dir_name = "test_basket_tmp_dir"
    tmp_basket_dir = test_pantry.set_up_basket(tmp_basket_dir_name)
    basket_path = test_pantry.upload_basket(tmp_basket_dir=tmp_basket_dir)

    basket = Basket(
        Path(basket_path),
        file_system=test_pantry.file_system
    )

    # Save the original basket supplement as a dictionary.
    original_supplement = basket.get_supplement()
    supplement_path = basket.supplement_path

    # Manually replace the Supplement file.
    test_pantry.file_system.rm(supplement_path)
    with test_pantry.file_system.open(supplement_path, "w") as outfile:
        json.dump({"junk": "b"}, outfile)

    # Supplement should already be cached and the new copy shouldn't be read.
    supplement = basket.get_supplement()
    assert supplement == {
        "upload_items": original_supplement["upload_items"],
        "integrity_data": original_supplement["integrity_data"],
    }


def test_basket_get_metadata(test_pantry):
    """Test that the get_metadata function returns the expected values."""
    metadata_in = {"test": 1}

    # Create a temporary basket with a test file, and upload it.
    tmp_basket_dir_name = "test_basket_tmp_dir"
    tmp_basket_dir = test_pantry.set_up_basket(tmp_basket_dir_name)
    basket_path = test_pantry.upload_basket(
        tmp_basket_dir, metadata=metadata_in
    )

    basket = Basket(
        Path(basket_path),
        file_system=test_pantry.file_system
    )

    # Check get_metadata returns the same values used during the upload.
    metadata = basket.get_metadata()
    assert metadata_in == metadata


def test_basket_get_metadata_cached(test_pantry):
    """Test that the get_metadata function retrieves cached copies of a
    basket's metadata.
    """

    metadata_in = {"test": 1}

    # Create a temporary basket with a test file, and upload it.
    tmp_basket_dir_name = "test_basket_tmp_dir"
    tmp_basket_dir = test_pantry.set_up_basket(tmp_basket_dir_name)
    basket_path = test_pantry.upload_basket(
        tmp_basket_dir, metadata=metadata_in
    )

    basket = Basket(
        Path(basket_path),
        file_system=test_pantry.file_system
    )

    # Save the original basket metadata as a dictionary.
    metadata = basket.get_metadata()
    metadata_path = basket.metadata_path

    # Manually replace the metadata file
    test_pantry.file_system.rm(metadata_path)
    with test_pantry.file_system.open(metadata_path, "w") as outfile:
        json.dump({"junk": "b"}, outfile)

    # Metadata should already be cached and the new copy shouldn't be read.
    metadata = basket.get_metadata()
    assert metadata_in == metadata


def test_basket_get_metadata_none(test_pantry):
    """Test that get_metadata returns None when no metadata was uploaded."""
    # Create a temporary basket with a test file, and upload it.
    tmp_basket_dir_name = "test_basket_tmp_dir"
    tmp_basket_dir = test_pantry.set_up_basket(tmp_basket_dir_name)
    basket_path = test_pantry.upload_basket(tmp_basket_dir)

    basket = Basket(
        Path(basket_path),
        file_system=test_pantry.file_system
    )
    metadata = basket.get_metadata()

    # No metadata was added to the upload, so it should be None.
    assert metadata is None


def test_basket_ls(test_pantry):
    """Test that the basket ls function returns the expected values."""
    # Create a temporary basket with a test file, and upload it.
    tmp_basket_dir_name = "test_basket_tmp_dir"
    tmp_basket_dir = test_pantry.set_up_basket(tmp_basket_dir_name)
    basket_path = test_pantry.upload_basket(tmp_basket_dir)

    basket = Basket(
        Path(basket_path),
        file_system=test_pantry.file_system
    )

    uploaded_dir_path = f"{basket_path}/{tmp_basket_dir_name}"
    assert basket.ls()[0].endswith(uploaded_dir_path)


def test_basket_ls_relpath(test_pantry):
    """Test that the basket ls function works when using relative paths."""
    # Create a temporary basket with a test file, and upload it.
    tmp_basket_dir_name = "test_basket_tmp_dir"
    tmp_basket_dir = test_pantry.set_up_basket(tmp_basket_dir_name)
    basket_path = test_pantry.upload_basket(tmp_basket_dir)

    basket = Basket(
        Path(basket_path),
        file_system=test_pantry.file_system
    )

    uploaded_file_path = f"{basket_path}/{tmp_basket_dir_name}/test.txt"
    assert basket.ls(Path(tmp_basket_dir_name))[0].endswith(uploaded_file_path)


def test_basket_ls_relpath_period(test_pantry):
    """Test that the basket ls function works when using the relative path '.'
    """

    # Create a temporary basket with a test file, and upload it.
    tmp_basket_dir_name = "test_basket_tmp_dir"
    tmp_basket_dir = test_pantry.set_up_basket(tmp_basket_dir_name)
    basket_path = test_pantry.upload_basket(tmp_basket_dir)

    basket = Basket(
        Path(basket_path),
        file_system=test_pantry.file_system
    )

    uploaded_dir_path = f"{basket_path}/{tmp_basket_dir_name}"
    assert basket.ls(".")[0].endswith(uploaded_dir_path)


def test_basket_ls_is_pathlike(test_pantry):
    """Test that the basket ls function only works with the expected
    value types.
    """

    # Create a temporary basket with a test file, and upload it.
    tmp_basket_dir_name = "test_basket_tmp_dir"
    tmp_basket_dir = test_pantry.set_up_basket(tmp_basket_dir_name)
    basket_path = test_pantry.upload_basket(tmp_basket_dir)

    basket = Basket(
        Path(basket_path),
        file_system=test_pantry.file_system
    )

    with pytest.raises(
        TypeError, match="expected str, bytes or os.PathLike object, not int"
    ):
        basket.ls(1)


def test_basket_ls_after_find(test_pantry):
    """The s3fs.S3FileSystem.ls() func is broken after running {}.find().

    This function is primarily to test s3fs file systems, but local
    file systems should yield the same results.

    s3fs.S3FileSystem.find() function is called during index creation. The
    solution to this problem is to ensure Basket.ls() uses the argument
    refresh=True when calling s3fs.ls(). This ensures that cached results
    from s3fs.find() (which is called during create_index_from_fs() and do not
    include directories) do not affect the s3fs.ls() function used to enable
    the Basket.ls() function.
    """

    tmp_basket_dir_name = "test_basket_temp_dir"
    tmp_basket_dir = test_pantry.set_up_basket(tmp_basket_dir_name)
    tmp_basket_dir = test_pantry.add_lower_dir_to_temp_basket(tmp_basket_dir)
    basket_path = test_pantry.upload_basket(tmp_basket_dir=tmp_basket_dir)

    # Create index on pantry
    create_index_from_fs(test_pantry.pantry_path, test_pantry.file_system)

    # Run find in case index creation changes
    test_pantry.file_system.find(test_pantry.pantry_path)

    # Set up basket
    test_basket = Basket(
        basket_path,
        file_system=test_pantry.file_system
    )

    expected_base_dir_paths = [
        os.path.join(basket_path, tmp_basket_dir_name, i)
        for i in ["nested_dir", "test.txt"]
    ]
    expected_base_dir_paths.sort()  # Sort to zip in same order

    ls_test = test_basket.ls(tmp_basket_dir_name)
    ls_test.sort()

    # Get the actual base dir paths (essentially stripping any FS specific
    # prefixes or conventions, ie in local file systems, the path to where
    # the script was called might be prepended, clean up stuff like that here)
    actual_bdp = [
        x.endswith(z)
        for x, z in zip(ls_test, expected_base_dir_paths, strict=True)
    ]

    # Check false is not in actual_bdp--which is a list of booleans that
    # indicates if the indices match.
    assert False not in actual_bdp


def test_basket_init_from_uuid(test_pantry):
    """Test that a basket can be successfully initialized from a UUID.
    """

    # Put basket in the temporary pantry
    tmp_basket_dir_one = test_pantry.set_up_basket("basket_one")
    uuid = "0000"
    test_pantry.upload_basket(tmp_basket_dir=tmp_basket_dir_one, uid=uuid)
    pantry = Pantry(IndexPandas,
                    pantry_path=test_pantry.pantry_path,
                    file_system=test_pantry.file_system)
    pantry.index.generate_index()
    test_basket = Basket(
        basket_address=uuid,
        pantry=pantry,
    )
    assert test_basket.ls("basket_one")[0].endswith(
        f"{test_pantry.pantry_path}/test_basket/0000/basket_one/test.txt"
    )


def test_basket_init_fails_if_uuid_does_not_exist(test_pantry):
    """Test that an error is raised when trying to initialize a basket using a
    UUID that does not have an associated basket.
    """

    # Put basket in the temporary pantry
    tmp_basket_dir_one = test_pantry.set_up_basket("basket_one")
    uuid = "0000"
    bad_uuid = "a bad uuid"
    test_pantry.upload_basket(tmp_basket_dir=tmp_basket_dir_one, uid=uuid)
    pantry = Pantry(IndexPandas,
                    pantry_path=test_pantry.pantry_path,
                    file_system=test_pantry.file_system)
    pantry.index.generate_index()
    with pytest.raises(ValueError, match=f"Basket does not exist: {bad_uuid}"):
        Basket(
            basket_address=bad_uuid,
            pantry=pantry,
        )


def test_basket_pantry_name_does_not_exist(test_pantry):
    """Test than an error is raised when trying to initialize a basket using a
    UUID, but using a bucket name that does not exist.
    """

    # Put basket in the temporary bucket
    tmp_basket_dir_one = test_pantry.set_up_basket("basket_one")
    uuid = "0000"
    test_pantry.upload_basket(tmp_basket_dir=tmp_basket_dir_one, uid=uuid)

    pantry_path = "the wrong pantry 007"
    error_msg = f'Invalid pantry Path. Pantry does not exist at: {pantry_path}'
    with pytest.raises(ValueError, match=error_msg):
        pantry = Pantry(IndexPandas,
                pantry_path=pantry_path,
                file_system=test_pantry.file_system)
        pantry.index.generate_index()


def test_basket_from_uuid_with_many_baskets(test_pantry):
    """Test that many baskets can be initialized using UUIDs."""
    # Set up ten baskets
    for uuid in range(10):
        uuid = str(uuid)
        tmp_basket_dir = test_pantry.set_up_basket(f"temp_basket_{uuid}")
        test_pantry.upload_basket(tmp_basket_dir=tmp_basket_dir, uid=uuid)

    pantry = Pantry(IndexPandas,
                    pantry_path=test_pantry.pantry_path,
                    file_system=test_pantry.file_system)
    pantry.index.generate_index()
    test_basket = Basket(
        basket_address=uuid,
        pantry=pantry,
    )
    assert test_basket.ls(f"temp_basket_{uuid}")[0].endswith(
        f"{test_pantry.pantry_path}/test_basket/{uuid}"
        f"/temp_basket_{uuid}/test.txt"
    )


def test_basket_correct_weave_version_member_variable(test_pantry):
    """Test that basket has the correct weave version as a member variable
    """
    tmp_basket_dir = test_pantry.set_up_basket("basket_one")
    basket_path = test_pantry.upload_basket(tmp_basket_dir=tmp_basket_dir)

    basket = Basket(
        Path(basket_path),
        pantry_path=test_pantry.pantry_path,
        file_system=test_pantry.file_system
    )

    assert basket.weave_version == weave_version


def test_basket_check_member_variables(test_pantry):
    """Check that you can access member variables in the basket"""
    # Upload a basket
    tmp_basket_dir = test_pantry.set_up_basket("basket")
    uuid = "0000"
    basket_path = test_pantry.upload_basket(tmp_basket_dir=tmp_basket_dir,
                                            uid=uuid)

    pantry = Pantry(IndexPandas,
                    pantry_path=test_pantry.pantry_path,
                    file_system=test_pantry.file_system)
    pantry.index.generate_index()
    my_basket = Basket(
        basket_address=uuid,
        pantry=pantry,
    )

    # Open the manifest to get the file system data
    manifest_path = os.path.join(basket_path, "basket_manifest.json")

    with test_pantry.file_system.open(manifest_path, "rb") as file:
        manifest_dict = json.load(file)

    # Validate the basket object's member variables match the file system data
    assert manifest_dict["uuid"] == my_basket.uuid
    assert manifest_dict["upload_time"] == my_basket.upload_time
    assert manifest_dict["parent_uuids"] == my_basket.parent_uuids
    assert manifest_dict["basket_type"] == my_basket.basket_type
    assert manifest_dict["label"] == my_basket.label
    assert manifest_dict["weave_version"] == weave_version
    assert my_basket.address.endswith(basket_path)
    assert test_pantry.file_system.__class__.__name__ == my_basket.storage_type


def test_basket_to_pandas_df(test_pantry):
    """Check that to_pandas_df returns the proper dataframe for the basket"""
    # Upload a basket
    tmp_basket_dir = test_pantry.set_up_basket("basket")
    uuid = "0000"
    basket_path = test_pantry.upload_basket(tmp_basket_dir=tmp_basket_dir,
                                            uid=uuid)

    pantry = Pantry(IndexPandas,
                    pantry_path=test_pantry.pantry_path,
                    file_system=test_pantry.file_system)
    pantry.index.generate_index()
    my_basket = Basket(
        basket_address=uuid,
        pantry=pantry,
    )

    basket_df = my_basket.to_pandas_df()

    # Open the manifest to get the file system data
    manifest_path = os.path.join(basket_path, "basket_manifest.json")

    with test_pantry.file_system.open(manifest_path, "rb") as file:
        manifest_dict = json.load(file)

    # Collect the data and build a dataframe for testing
    data = [manifest_dict["uuid"],
            manifest_dict["upload_time"],
            manifest_dict["parent_uuids"],
            manifest_dict["basket_type"],
            manifest_dict["label"],
            manifest_dict["weave_version"],
            basket_path,
            test_pantry.file_system.__class__.__name__]

    columns = ["uuid", "upload_time", "parent_uuids",
               "basket_type", "label", "weave_version",
               "address", "storage_type"]

    answer_df = pd.DataFrame(data=[data], columns=columns)

    # Addresses can have different prefixes in the paths, check that it ends
    # with the correct relative path
    assert basket_df["address"][0].endswith(answer_df["address"][0])

    # Drop the addresses that could be slightly different and compare the rest
    # of the dataframe
    basket_df.drop(columns="address", inplace=True)
    answer_df.drop(columns="address", inplace=True)
    assert basket_df.equals(answer_df)

def test_basket_time_is_utc(test_pantry):
    """Make sure time data is in UTC format"""
    # Upload a basket
    uuid = "0000"
    tmp_basket_dir = test_pantry.set_up_basket("basket")
    basket_path = test_pantry.upload_basket(tmp_basket_dir=tmp_basket_dir,
                                            uid=uuid)

    manifest_path = os.path.join(basket_path, "basket_manifest.json")

    with test_pantry.file_system.open(manifest_path, "rb") as file:
        manifest_dict = json.load(file)

    # Regex to match ISO 8601 (UTC)
    regex = (
        r"^(-?(?:[1-9][0-9]*)?[0-9]{4})-(1[0-2]|0[1-9])-"
        r"(3[01]|0[1-9]|[12][0-9])T(2[0-3]|[01][0-9]):"
        r"([0-5][0-9]):([0-5][0-9])(\.[0-9]+)?"
        r"(Z|[+-](?:2[0-3]|[01][0-9]):[0-5][0-9])?$"
    )

    match_iso8601 = re.compile(regex).match
    assert match_iso8601(str(manifest_dict['upload_time'])) is not None



"""Contains scripts concerning MongoDB functionality."""

import pandas as pd
# Try-Except required to make pymongo an optional dependency.

# Ignore pylint duplicate code. Code here is used to explicitly show pymongo is
# an optional dependency. Duplicate code is found in config.py (where pymongo
# is actually imported)
# pylint: disable-next=duplicate-code
try:
    # For the sake of explicitly showing that pymongo is optional, import
    # pymongo here, even though it is not currently used in this file.
    # Pylint ignore the next unused-import pylint warning.
    # Also inline ruff ignore unused import (F401)
    # pylint: disable-next=unused-import
    import pymongo # noqa: F401
except ImportError:
    _HAS_PYMONGO = False
else:
    _HAS_PYMONGO = True

from .basket import Basket
from .config import get_file_system, get_mongo_db


def load_mongo(index_table, collection="metadata", **kwargs):
    """Load metadata from baskets into the mongo database.

       A metadata.json is created in baskets when the metadata
       field is provided upon upload. This metadata is added to the
       Mongo database when invoking load_mongo. UUID, and basket_type
       from the index_table are also added to Mongo for referrence
       back to the datasource.

        Parameters
        ----------
        index_table: dataframe
            Weave index dataframe fetched using the Index class.
            The dataframe must include the following columns:
               uuid
               basket_type
               address
        collection: str (default="metadata")
            Metadata will be added to the Mongo collection specified.
        **file_system: fsspec object (required)
            The file system to retrieve the baskets' metadata from.
        """
    if not _HAS_PYMONGO:
        raise ImportError("Missing Dependency. The package 'pymongo' "
                          "is required to use this function")

    if not isinstance(index_table, pd.DataFrame):
        raise TypeError("Invalid datatype for index_table: "
                        "must be Pandas DataFrame")

    if not isinstance(collection, str):
        raise TypeError("Invalid datatype for collection: "
                        "must be a string")

    required_columns = ["uuid", "basket_type", "address"]

    for required_column in required_columns:
        if required_column not in index_table.columns.values.tolist():
            raise ValueError("Invalid index_table: missing "
                             f"{required_column} column")

    file_system = kwargs.get("file_system", get_file_system())
    database = get_mongo_db().mongo_metadata

    for _, row in index_table.iterrows():
        basket = Basket(row["address"], file_system=file_system)
        metadata = basket.get_metadata()
        if metadata is None:
            continue
        manifest = basket.get_manifest()
        mongo_metadata = {}
        mongo_metadata["uuid"] = manifest["uuid"]
        mongo_metadata["basket_type"] = manifest["basket_type"]
        mongo_metadata.update(metadata)

        # If the UUID already has metadata loaded in MongoDB,
        # the metadata should not be loaded to MongoDB again.
        if 0 == database[
            collection
        ].count_documents({"uuid": manifest["uuid"]}):
            database[collection].insert_one(mongo_metadata)



"""Pytest tests for the sqlite index."""
import os

import pytest
import s3fs
from fsspec.implementations.local import LocalFileSystem

from weave.pantry import Pantry
from weave.index.index_sqlite import IndexSQLite
from weave.tests.pytest_resources import PantryForTest


###############################################################################
#                      Pytest Fixtures Documentation:                         #
#            https://docs.pytest.org/en/7.3.x/how-to/fixtures.html            #
#                                                                             #
#                  https://docs.pytest.org/en/7.3.x/how-to/                   #
#          fixtures.html#teardown-cleanup-aka-fixture-finalization            #
#                                                                             #
#  https://docs.pytest.org/en/7.3.x/how-to/fixtures.html#fixture-parametrize  #
###############################################################################

# Pylint doesn't like redefining the test fixture here from
# test_basket, but this is the right way to do it if at some
# point in the future the two need to be differentiated.
# pylint: disable=duplicate-code

s3fs = s3fs.S3FileSystem(
    client_kwargs={"endpoint_url": os.environ["S3_ENDPOINT"]}
)
local_fs = LocalFileSystem()


# Test with two different fsspec file systems (above).
@pytest.fixture(
    name="test_pantry",
    params=[s3fs, local_fs],
    ids=["S3FileSystem", "LocalFileSystem"],
)
def fixture_test_pantry(request, tmpdir):
    """Sets up test pantry for the tests"""
    file_system = request.param
    test_pantry = PantryForTest(tmpdir, file_system)
    yield test_pantry
    test_pantry.cleanup_pantry()


def test_index_two_pantries_with_same_name(test_pantry):
    """Validate that 2 pantry objects with the same basename have their
    own index.
    """
    pantry_1_path = os.path.join(test_pantry.pantry_path, "test-pantry-1")
    pantry_2_path = os.path.join(test_pantry.pantry_path,
                                 "test",
                                 "test-pantry-1")

    test_pantry.file_system.mkdir(pantry_1_path)
    test_pantry.file_system.mkdir(pantry_2_path)

    tmp_txt_file = test_pantry.tmpdir.join("test.txt")
    tmp_txt_file.write("this is a test")

    upload_path_1 = os.path.join(pantry_1_path, "text.txt")
    upload_path_2 = os.path.join(pantry_2_path, "text.txt")

    # Must upload a file because Minio will remove empty directories
    test_pantry.file_system.upload(str(tmp_txt_file.realpath()), upload_path_1)
    test_pantry.file_system.upload(str(tmp_txt_file.realpath()), upload_path_2)

    # Make the Pantries.
    pantry_1 = Pantry(
        IndexSQLite,
        pantry_path=pantry_1_path,
        file_system=test_pantry.file_system,
    )

    pantry_2 = Pantry(
        IndexSQLite,
        pantry_path=pantry_2_path,
        file_system=test_pantry.file_system,
    )

    # Because the two Pantry Objects have the same name, or path basename,
    # the db files must be different.
    assert pantry_1.index.db_path != pantry_2.index.db_path

    # Remove the .db files that are not cleaned up with 'test_pantry'
    os.remove(pantry_1.index.db_path)
    os.remove(pantry_2.index.db_path)


def test_index_uploaded_basket_not_found_in_another_index(test_pantry):
    """Validate that a basket uploaded to one pantry does not show up in
    another pantry.
    """
    pantry_1_path = os.path.join(test_pantry.pantry_path, "test-pantry-1")
    pantry_2_path = os.path.join(test_pantry.pantry_path, "test-pantry-2")

    test_pantry.file_system.mkdir(pantry_1_path)
    test_pantry.file_system.mkdir(pantry_2_path)

    tmp_txt_file = test_pantry.tmpdir.join("test.txt")
    tmp_txt_file.write("this is a test")

    upload_path_1 = os.path.join(pantry_1_path, "text.txt")
    upload_path_2 = os.path.join(pantry_2_path, "text.txt")

    # Must upload a file because Minio will remove empty directories
    test_pantry.file_system.upload(str(tmp_txt_file.realpath()), upload_path_1)
    test_pantry.file_system.upload(str(tmp_txt_file.realpath()), upload_path_2)

    # Make the Pantries.
    pantry_1 = Pantry(
        IndexSQLite,
        pantry_path=pantry_1_path,
        file_system=test_pantry.file_system,
    )

    pantry_2 = Pantry(
        IndexSQLite,
        pantry_path=pantry_2_path,
        file_system=test_pantry.file_system,
    )

    # Upload a basket to each pantry.
    pantry_1.upload_basket(
        upload_items=[{"path":str(tmp_txt_file.realpath()), "stub":False}],
        basket_type="test-basket",
        unique_id="0000",
    )
    pantry_2.upload_basket(
        upload_items=[{"path":str(tmp_txt_file.realpath()), "stub":False}],
        basket_type="test-basket",
        unique_id="0001",
    )

    # Check that the indices are not the same
    pantry_1_index = pantry_1.index.to_pandas_df()
    pantry_2_index = pantry_2.index.to_pandas_df()

    # Validate that each basket was uploaded to the proper pantry, and
    # not the other.
    assert not pantry_1_index.equals(pantry_2_index)
    assert "0000" in pantry_1_index["uuid"].values
    assert "0000" not in pantry_2_index["uuid"].values
    assert "0001" in pantry_2_index["uuid"].values
    assert "0001" not in pantry_1_index["uuid"].values

    # Remove the .db files that are not cleaned up with 'test_pantry'
    os.remove(pantry_1.index.db_path)
    os.remove(pantry_2.index.db_path)



"""This class builds the user-facing Index class. It pulls from the _Index 
class which uses Pandas as it's backend to build and interface with the on disk
Index baskets.
"""

import json
import os

from .basket import Basket
from .config import get_file_system
from .index.create_index import create_index_from_fs
from .index.index_abc import IndexABC
from .upload import UploadBasket
from .validate import validate_pantry

class Pantry():
    """Facilitate user interaction with the index of a Weave data warehouse.
    """

    def __init__(self, index: IndexABC, pantry_path="weave-test", **kwargs):
        """Initialize Pantry object.

        A pantry is a collection of baskets. This class facilitates the upload,
        retrieval, and deletion of baskets within a file system. It uses and
        updates an index to track the contents within the pantry for rapid
        operations.

        Parameters:
        -----------
        index: IndexABC
            The concrete implementation of an IndexABC. This is used to track
            the contents within the pantry.
        pantry_path: str (default="weave-test")
            Name of the pantry this object is associated with.
        **file_system: fsspec object (optional)
            The fsspec object which hosts the pantry we desire to index.
            If file_system is None, then the default fs is retrieved from the
            config.
        """

        self.file_system = kwargs.pop("file_system", get_file_system())
        if not self.file_system.exists(pantry_path):
            raise ValueError(
                f"Invalid pantry Path. Pantry does not exist at: "
                f"{pantry_path}"
            )

        self.pantry_path = str(pantry_path)
        self.load_metadata()

        self.index = index(file_system=self.file_system,
                           pantry_path=self.pantry_path,
                           metadata=self.metadata['index_metadata'],
                           **kwargs
        )
        self.metadata['index_metadata'] = self.index.generate_metadata()

    def validate_path_in_pantry(self, path):
        """Validate the given path is within the pantry.

        Check 1: Ensure the path begins with the pantry path.
        Check 2: Ensure the ".." navigation command is not used.

        Parameters:
        -----------
        path: str
            Path to verify.
        """
        valid = path.startswith(self.pantry_path + os.path.sep)
        if valid:
            bad_str = os.path.sep + ".." + os.path.sep
            valid = bad_str not in path

        if not valid:
            raise ValueError(
                f"Attempting to access basket outside of pantry: {path}"
            )

    def load_metadata(self):
        """Load pantry metadata from pantry_metadata.json."""
        self.metadata_path = os.path.join(
                                self.pantry_path,'pantry_metadata.json'
        )
        if self.file_system.exists(self.metadata_path):
            with self.file_system.open(self.metadata_path, "rb") as file:
                self.metadata = json.load(file)
        else:
            self.metadata = {}
        if 'index_metadata' not in self.metadata:
            self.metadata['index_metadata'] = {}

    def save_metadata(self):
        """Dump metadata to to pantry metadata file."""
        self.metadata['index_metadata'] = self.index.metadata
        with self.file_system.open(
                    self.metadata_path, "w", encoding="utf-8"
        ) as outfile:
            json.dump(self.metadata, outfile)

    def validate(self):
        """Convenient wrapper function to validate the pantry.

        Returns
        ----------
        A list of all invalid basket locations (will return an empty list if
        no warnings are raised)
        """

        return validate_pantry(self)

    def delete_basket(self, basket_address, **kwargs):
        """Deletes basket of given UUID or path.

        Note that the given basket will not be deleted if the basket is listed
        as the parent uuid for any of the baskets in the index.

        Parameters:
        -----------
        basket_address: str
            Argument can take one of two forms: either a path to the basket
            directory, or the UUID of the basket.
        **kwargs:
            Additional parameters to pass to the index
        """

        basket_address = str(basket_address)
        remove_item = self.index.get_rows(basket_address)

        if len(self.index.get_children(remove_item.iloc[0].uuid)) > 0:
            raise ValueError(
                f"The provided value for basket_uuid {basket_address} " +
                "is listed as a parent UUID for another basket. Please " +
                "delete that basket before deleting its parent basket."
            )

        self.validate_path_in_pantry(remove_item.iloc[0].address)
        self.index.untrack_basket(remove_item.iloc[0].address, **kwargs)
        self.file_system.rm(remove_item.iloc[0].address, recursive=True)

    def upload_basket(self, upload_items, basket_type, **kwargs):
        """Upload a basket to the same pantry referenced by the Index

        Parameters
        ----------
        upload_items : [dict]
            List of python dictionaries with the following schema:
            {
                'path': path to the file or folder being uploaded (string),
                'stub': true/false (bool)
            }
            'path' can be a file or folder to be uploaded. Every filename
            and folder name must be unique. If 'stub' is set to True, integrity
            data will be included without uploading the actual file or folder.
            Stubs are useful when original file source information is desired
            without uploading the data itself. This is especially useful when
            dealing with large files.
        basket_type: str
            Type of basket being uploaded.
        **parent_ids: [str] (optional)
            List of unique ids associated with the parent baskets
            used to derive the new basket being uploaded.
        **metadata: dict (optional)
            Python dictionary that will be written to metadata.json
            and stored in the basket in upload file_system.
        **label: str (optional)
            Optional user friendly label associated with the basket.
        """

        parent_ids = kwargs.pop("parent_ids", [])
        metadata = kwargs.pop("metadata", {})
        label = kwargs.pop("label", "")

        up_dir = UploadBasket(
            upload_items=upload_items,
            basket_type=basket_type,
            file_system=self.file_system,
            pantry_path=self.pantry_path,
            parent_ids=parent_ids,
            metadata=metadata,
            label=label,
            **kwargs,
        ).get_upload_path()

        single_indice_index = create_index_from_fs(up_dir, self.file_system)
        self.index.track_basket(single_indice_index)
        return single_indice_index

    def get_basket(self, basket_address):
        """Retrieves a basket of given UUID or path.

        Parameters
        ----------
        basket_address: str
            Argument can take one of two forms: either a path to the Basket
            directory, or the UUID of the basket.

        Returns
        ----------
        The Basket object associated with the given UUID or path.
        """

        # Create a Basket from the given address, and the index's file_system
        # and pantry name. Basket will catch invalid inputs and raise
        # appropriate errors.
        row = self.index.get_rows(basket_address)
        if len(row) == 0:
            raise ValueError(f"Basket does not exist: {basket_address}")
        self.validate_path_in_pantry(row.iloc[0].address)
        return Basket(row.iloc[0].address, pantry=self)



"""Resources for use in pytest."""

import json
import os

from weave.upload import UploadBasket


def file_path_in_list(search_path, search_list):
    """Check if a file path is in a list (of file paths).

    Parameters
    ----------
    search_path: str
        The file path being searched for.
    search_list: [str]
        A list of strings (presumably file paths).

    Returns
    ----------
    bool: True if any of the file paths in the search list end with the
    file path being searched for, otherwise False.

    This allows us to determine if a file is in the list regardless of file
    system dependent prefixes such as /home/user/ which can usually be ignored.
    For example, given a list of file paths:
    ['/home/user/data/file.txt', '/home/user/data/this/is/test.txt']
    and the function is searching for 'data/file.txt',
    the function will return True as the file exists.
    """

    search_path = str(search_path)
    for file_path in search_list:
        if str(file_path).endswith(search_path):
            return True

    return False


class PantryForTest:
    """Handles resources for much of weave testing."""
    def __init__(self, tmpdir, file_system, pantry_path=None):
        self.tmpdir = tmpdir
        self.file_system = file_system
        self.pantry_path = pantry_path

        if self.pantry_path is None:
            self.pantry_path = (
                "pytest-temp-pantry"
                f"{os.environ.get('WEAVE_PYTEST_SUFFIX', '')}"
            )
        self.basket_list = []
        self._set_up_pantry()

    def _set_up_pantry(self):
        """Create a temporary pantry for testing purposes."""
        try:
            self.file_system.mkdir(self.pantry_path)
        except FileExistsError:
            self.cleanup_pantry()
            self._set_up_pantry()

    def set_up_basket(
        self, tmp_dir_name, file_name="test.txt", file_content="This is a test"
    ):
        """Create a temporary (local) basket, with a single text file."""
        tmp_basket_dir = self.tmpdir.mkdir(tmp_dir_name)
        tmp_basket_txt_file = tmp_basket_dir.join(file_name)

        if file_name[file_name.rfind(".") :] == ".json":
            with open(tmp_basket_txt_file, "w", encoding="utf-8") as outfile:
                json.dump(file_content, outfile)
        else:
            tmp_basket_txt_file.write(file_content)

        return tmp_basket_dir

    def add_lower_dir_to_temp_basket(self, tmp_basket_dir):
        """Add a nested directory inside the temporary basket."""
        nested_dir = tmp_basket_dir.mkdir("nested_dir")
        nested_dir.join("another_test.txt").write("more test text")
        return tmp_basket_dir

    def upload_basket(
        self, tmp_basket_dir, uid="0000", basket_type="test_basket", **kwargs
    ):
        """Upload a temporary (local) basket to the S3 test pantry."""
        upload_items = [
            {"path": str(tmp_basket_dir.realpath()), "stub": False}
        ]

        # Ignore pylint duplicate code. Code here is required to ensure proper
        # upload, and is similar to the pantry upload call--obviously.
        # pylint: disable-next=duplicate-code
        up_dir = UploadBasket(
            upload_items=upload_items,
            basket_type=basket_type,
            file_system=self.file_system,
            pantry_path=self.pantry_path,
            unique_id=uid,
            **kwargs,
        ).get_upload_path()
        return up_dir

    def cleanup_pantry(self):
        """Delete the temporary test pantry, including any uploaded baskets."""
        self.file_system.rm(self.pantry_path, recursive=True)

# This class is to facilitate creating and deleting indices for tests.
# No other functionality required, but it is needed for pytest
# pylint: disable-next=too-few-public-methods
class IndexForTest:
    """Creates an index for testing."""
    def __init__(self, index_constructor, file_system, pantry_path=None):
        self.file_system = file_system
        self.pantry_path = pantry_path

        if self.pantry_path is None:
            self.pantry_path = (
                "pytest-temp-pantry"
                f"{os.environ.get('WEAVE_PYTEST_SUFFIX', '')}"
            )

        # This is only used in the sqlite implementation and will not affect
        # other implmentations.
        self.db_path = f"{self.pantry_path}.db"

        self.index = index_constructor(
            file_system=self.file_system,
            pantry_path=self.pantry_path,
            db_path=self.db_path,
        )

    def cleanup_index(self):
        """Clean up any artifacts created by index implementations."""
        # Remove SQLite db file.
        if os.path.exists(self.db_path):
            os.remove(self.db_path)



"""Pytest tests for the index directory."""

import os
import re
import warnings
from datetime import datetime, timedelta, timezone

import numpy as np
import pandas as pd
import pytest
import s3fs
from fsspec.implementations.local import LocalFileSystem

import weave
from weave import IndexSQLite
from weave.index.index_pandas import IndexPandas
from weave.index.create_index import create_index_from_fs
from weave.tests.pytest_resources import PantryForTest, IndexForTest


###############################################################################
#                      Pytest Fixtures Documentation:                         #
#            https://docs.pytest.org/en/7.3.x/how-to/fixtures.html            #
#                                                                             #
#                  https://docs.pytest.org/en/7.3.x/how-to/                   #
#          fixtures.html#teardown-cleanup-aka-fixture-finalization            #
#                                                                             #
#  https://docs.pytest.org/en/7.3.x/how-to/fixtures.html#fixture-parametrize  #
###############################################################################

# This module is long and has many tests. Pylint is complaining that it is too
# long. This isn't neccesarily a bad thing, as the alternative would be to
# write the tests continuing in a different script, which is unnecessarily
# complex. Disabling this warning for this script.
# pylint: disable=too-many-lines

# Pylint doesn't like redefining the test fixture here from
# test_basket, but this is the right way to do this in case at some
# point in the future there is a need to differentiate the two.
# pylint: disable=duplicate-code


# Create fsspec objects to be tested, and add to file_systems list.
s3fs = s3fs.S3FileSystem(
    client_kwargs={"endpoint_url": os.environ["S3_ENDPOINT"]}
)
local_fs = LocalFileSystem()
file_systems = [s3fs, local_fs]

# Create Index CONSTRUCTORS of Indexes to be tested, and add to indexes list.
indexes = [IndexSQLite, IndexPandas]
indexes_ids = ["SQLite", "Pandas"]

# Create combinations of the above parameters to pass into the fixture..
params = []
params_ids = []
for iter_file_system in file_systems:
    for iter_index, iter_index_id in zip(indexes, indexes_ids):
        params.append((iter_file_system, iter_index))
        params_ids.append(f"{iter_file_system.__class__.__name__}-"
                          f"-{iter_index_id}")


@pytest.fixture(name="test_pantry", params=params, ids=params_ids)
def fixture_test_pantry(request, tmpdir):
    """Sets up test pantry for the tests."""
    file_system = request.param[0]
    pantry_path = (
        "pytest-temp-pantry" f"{os.environ.get('WEAVE_PYTEST_SUFFIX', '')}"
    )

    test_pantry = PantryForTest(tmpdir, file_system, pantry_path=pantry_path)

    index_constructor = request.param[1]
    test_index = IndexForTest(
        index_constructor=index_constructor,
        file_system=file_system,
        pantry_path=pantry_path
    )
    index = test_index.index

    yield test_pantry, index
    test_pantry.cleanup_pantry()
    test_index.cleanup_index()


@pytest.fixture(name="test_index_only", params=indexes, ids=indexes_ids)
def fixture_test_index_only(request):
    """Sets up only the index for a test (USES DEFAULT FILE_SYSTEM ARGS).

    Use this fixture for tests that DO NOT manipulate the pantry (ie
    type checking tests, etc.)
    """
    index_constructor = request.param
    file_system = weave.config.get_file_system()
    pantry_path = (
        "pytest-temp-pantry" f"{os.environ.get('WEAVE_PYTEST_SUFFIX', '')}"
    )

    test_index = IndexForTest(
        index_constructor=index_constructor,
        file_system=file_system,
        pantry_path=pantry_path
    )
    index = test_index.index

    yield index
    test_index.cleanup_index()


def test_index_abc_builtin_len_works(test_pantry):
    """Test IndexABC builtin __len__ returns number of baskets being tracked.
    """
    # Unpack the test_pantry into two variables for the pantry and index.
    test_pantry, ind = test_pantry

    # Put basket in the temporary pantry.
    tmp_basket_dir_one = test_pantry.set_up_basket("basket_one")
    test_pantry.upload_basket(tmp_basket_dir=tmp_basket_dir_one, uid="0001")
    tmp_basket_dir_one = test_pantry.set_up_basket("basket_two")
    test_pantry.upload_basket(tmp_basket_dir=tmp_basket_dir_one, uid="0002")
    tmp_basket_dir_one = test_pantry.set_up_basket("basket_three")
    test_pantry.upload_basket(tmp_basket_dir=tmp_basket_dir_one, uid="0003")

    # Generate the index.
    ind.generate_index()
    assert len(ind) == 3, f"Builtin len returns inaccurate value: {len(ind)}"


def test_index_abc_builtin_str_works(test_index_only):
    """Test IndexABC builtin __str__ returns str of the concrete class name."""
    ind = test_index_only
    assert ind.__class__.__name__ == str(ind), (
        f"Buildin str returns incorrect value: {str(ind)}"
    )


def test_index_abc_generate_metadata_returns_dict(test_index_only):
    """Test IndexABC generate_metadata returns a python dictionary."""
    ind = test_index_only

    metadata = ind.generate_metadata()
    assert (
        isinstance(metadata, dict)
    ), "Index.generate_metadata must return a dict."


def test_index_abc_generate_index_works(test_pantry):
    """Tests IndexABC generate_index uses the pantry fs to add to the index."""
    # Unpack the test_pantry into two variables for the pantry and index.
    test_pantry, ind = test_pantry

    # Generate the index.
    ind.generate_index()
    assert len(ind) == 0, "Incorrect number of elements in the index."

    # Put basket in the temporary pantry
    tmp_basket_dir_one = test_pantry.set_up_basket("basket_one")
    test_pantry.upload_basket(tmp_basket_dir=tmp_basket_dir_one, uid="0001")

    # Generate the index.
    ind.generate_index()
    assert len(ind) == 1, "Incorrect number of elements in the index."

    # Add another basket to the temporary pantry
    tmp_basket_dir_one = test_pantry.set_up_basket("basket_two")
    test_pantry.upload_basket(tmp_basket_dir=tmp_basket_dir_one, uid="0002")

    # Regenerate the index
    ind.generate_index()
    assert len(ind) == 2, "Incorrect number of elements in the index."


def test_index_abc_to_pandas_df_works(test_pantry):
    """Tests IndexABC to_pandas_df returns dataframe with proper values."""
    # Unpack the test_pantry into two variables for the pantry and index.
    test_pantry, ind = test_pantry

    # Put basket in the temporary pantry
    uuid, basket_type, label = "0001", "test_basket", "test_label"
    tmp_basket_dir_one = test_pantry.set_up_basket("basket_one")
    up_dir = test_pantry.upload_basket(
        tmp_basket_dir=tmp_basket_dir_one,
        uid=uuid,
        basket_type=basket_type,
        label=label
    )

    # Generate the index.
    ind.generate_index()
    ind_df = ind.to_pandas_df()

    # Check a pandas dataframe returns of the correct length.
    assert len(ind_df) == 1 and isinstance(ind_df, pd.DataFrame)

    # Check df columns are named correctly.
    assert (
        list(ind_df.columns) == weave.config.get_index_column_names()
    ), "Dataframe columns do not match"

    # Check values of basket are accurate
    assert (
        ind_df.iloc[0]["uuid"] == uuid and
        isinstance(ind_df.iloc[0]["upload_time"], datetime) and
        ind_df.iloc[0]["parent_uuids"] == [] and
        ind_df.iloc[0]["basket_type"] == basket_type and
        ind_df.iloc[0]["label"] == label and
        ind_df.iloc[0]["address"].endswith(up_dir) and
        ind_df.iloc[0]["storage_type"] == \
            test_pantry.file_system.__class__.__name__
    ), "Retrieved manifest values do not match."

    # Upload another basket
    uuid, basket_type, label = "0002", "test_basket", "test_label"
    parent_ids = ["0001"]
    tmp_basket_dir_one = test_pantry.set_up_basket("basket_two")
    up_dir = test_pantry.upload_basket(
        tmp_basket_dir=tmp_basket_dir_one,
        uid=uuid,
        basket_type=basket_type,
        parent_ids=parent_ids,
        label=label
    )

    # Generate the index.
    ind.generate_index()
    ind_df = ind.to_pandas_df()

    # Check a pandas dataframe returns of the correct length.
    assert len(ind_df) == 2 and isinstance(ind_df, pd.DataFrame)

    # Check df columns are named correctly.
    assert (
        list(ind_df.columns) == weave.config.get_index_column_names()
    ), "Dataframe columns do not match"

    # Check values of basket are accurate
    assert (
        ind_df.iloc[1]["uuid"] == uuid and
        isinstance(ind_df.iloc[1]["upload_time"], datetime) and
        ind_df.iloc[1]["parent_uuids"] == parent_ids and
        ind_df.iloc[1]["basket_type"] == basket_type and
        ind_df.iloc[1]["label"] == label and
        ind_df.iloc[1]["address"].endswith(up_dir) and
        ind_df.iloc[1]["storage_type"] == \
            test_pantry.file_system.__class__.__name__
    ), "Retrieved manifest values do not match."


def test_index_abc_track_basket_adds_single_basket(test_pantry):
    """Test IndexABC track_basket works when passing a single basket df."""
    # Unpack the test_pantry into two variables for the pantry and index.
    test_pantry, ind = test_pantry

    # Generate the index.
    ind.generate_index()
    assert len(ind) == 0, "Incorrect number of elements in the index."

    # Put basket in the temporary pantry
    uuid, basket_type, label = "0001", "test_basket", "test_label"
    tmp_basket_dir_one = test_pantry.set_up_basket("basket_one")
    up_dir = test_pantry.upload_basket(
        tmp_basket_dir=tmp_basket_dir_one,
        uid=uuid,
        basket_type=basket_type,
        label=label
    )

    single_indice_index = create_index_from_fs(
        up_dir,
        test_pantry.file_system
    )

    ind.track_basket(single_indice_index)
    ind_df = ind.to_pandas_df()
    assert len(ind) == 1, "Incorrect number of elements in the Index"

    # Check values of basket are accurate
    assert (
        ind_df.iloc[0]["uuid"] == uuid and
        isinstance(ind_df.iloc[0]["upload_time"], datetime) and
        ind_df.iloc[0]["parent_uuids"] == [] and
        ind_df.iloc[0]["basket_type"] == basket_type and
        ind_df.iloc[0]["label"] == label and
        ind_df.iloc[0]["address"].endswith(up_dir) and
        ind_df.iloc[0]["storage_type"] == \
            test_pantry.file_system.__class__.__name__
    ), "Retrieved manifest values do not match."


def test_index_abc_track_basket_adds_multiple_baskets(test_pantry):
    """Test IndexABC track_basket works when passing a multi-basket df."""
    # Unpack the test_pantry into two variables for the pantry and index.
    test_pantry, ind = test_pantry

    # Generate the index.
    ind.generate_index()

    # Put basket in the temporary pantry
    uuids = ["0001", "0002"]
    basket_type = "test_basket"
    test_label = "test_label"
    tmp_basket_dir_one = test_pantry.set_up_basket("basket_one")
    up_dir1 = test_pantry.upload_basket(
        tmp_basket_dir=tmp_basket_dir_one,
        uid=uuids[0],
        basket_type=basket_type,
        label=test_label
    )
    first_slice_df = create_index_from_fs(up_dir1, test_pantry.file_system)

    # Upload another basket
    parent_ids2 = ["0001"]
    tmp_basket_dir_two = test_pantry.set_up_basket("basket_two")
    up_dir2 = test_pantry.upload_basket(
        tmp_basket_dir=tmp_basket_dir_two,
        uid=uuids[1],
        basket_type=basket_type,
        parent_ids=parent_ids2,
        label=test_label
    )
    second_slice_df = create_index_from_fs(up_dir2, test_pantry.file_system)

    dual_slice_df = pd.concat([first_slice_df, second_slice_df])
    assert len(dual_slice_df) == 2, "Invalid dual_indice_index values."

    ind.track_basket(dual_slice_df)
    ind_df = ind.to_pandas_df()

    # Check a pandas dataframe returns of the correct length.
    assert (
        len(ind_df) == 2 and isinstance(ind_df, pd.DataFrame)
    ), "track_basket failed to add multiple items."

    # Check values of basket are accurate
    assert (
        ind_df.iloc[0]["uuid"] == uuids[0] and
        isinstance(ind_df.iloc[0]["upload_time"], datetime) and
        ind_df.iloc[0]["parent_uuids"] == [] and
        ind_df.iloc[0]["basket_type"] == basket_type and
        ind_df.iloc[0]["label"] == test_label and
        ind_df.iloc[0]["address"].endswith(up_dir1) and
        ind_df.iloc[0]["storage_type"] == \
            test_pantry.file_system.__class__.__name__
    ), "Retrieved manifest values do not match first record."

    assert (
        ind_df.iloc[1]["uuid"] == uuids[1] and
        isinstance(ind_df.iloc[1]["upload_time"], datetime) and
        ind_df.iloc[1]["parent_uuids"] == ["0001"] and
        ind_df.iloc[1]["basket_type"] == basket_type and
        ind_df.iloc[1]["label"] == test_label and
        ind_df.iloc[1]["address"].endswith(up_dir2) and
        ind_df.iloc[1]["storage_type"] == \
            test_pantry.file_system.__class__.__name__
    ), "Retrieved manifest values do not match second record."

    # Check df columns are named correctly.
    assert (
        list(ind_df.columns) == weave.config.get_index_column_names()
    ), "Dataframe columns do not match"


def test_index_abc_untrack_basket_removes_single_basket(test_pantry):
    """Test IndexABC untrack_basket works when passing a basket address."""
    # Unpack the test_pantry into two variables for the pantry and index.
    test_pantry, ind = test_pantry

    # Put basket in the temporary pantry.
    tmp_basket_dir_one = test_pantry.set_up_basket("basket_one")
    test_pantry.upload_basket(tmp_basket_dir=tmp_basket_dir_one, uid="0001")
    tmp_basket_dir_one = test_pantry.set_up_basket("basket_two")
    test_pantry.upload_basket(tmp_basket_dir=tmp_basket_dir_one, uid="0002")
    tmp_basket_dir_one = test_pantry.set_up_basket("basket_three")
    test_pantry.upload_basket(tmp_basket_dir=tmp_basket_dir_one, uid="0003")

    # Create the addresses to remove the last two baskets.
    untrack_uuid = "0002"
    untrack_path = os.path.join(test_pantry.pantry_path, "test_basket", "0003")

    # Generate the index.
    ind.generate_index()

    ind.untrack_basket(untrack_uuid)
    assert len(ind) == 2, "untrack_basket(uuid) failed to update index."

    ind.untrack_basket(untrack_path)
    assert len(ind) == 1, "untrack_basket(path) failed to update index."


def test_index_abc_untrack_basket_raises_warning(test_pantry):
    """Test IndexABC untrack_basket raises a warning when the address to be
    untracked wasn't tracked to begin with."""
    # Unpack the test_pantry into two variables for the pantry and index.
    test_pantry, ind = test_pantry

    # Put basket in the temporary pantry.
    tmp_basket_dir_one = test_pantry.set_up_basket("basket_one")
    test_pantry.upload_basket(tmp_basket_dir=tmp_basket_dir_one, uid="0001")

    untracked_uuid = "0009"
    untracked_path = os.path.join("test-pantry-that-isnt-real", "fake_basket")

    # Generate the index.
    ind.generate_index()

    # Test the warning is raised when untracking an already untracked UUID.
    with warnings.catch_warnings(record=True) as warn:
        ind.untrack_basket(untracked_uuid)
    warning_list = [warn[i].message for i in range(len(warn))]
    warning_1 = warning_list[0]
    assert warning_1.args[0] == (
        "Incomplete Request. Index could not untrack baskets, "
        "as some were not being tracked to begin with."
    )
    assert warning_1.args[1] == 1

    # Test the warning is raised when untracking an already untracked PATH.
    with warnings.catch_warnings(record=True) as warn:
        ind.untrack_basket(untracked_path)
    warning_list = [warn[i].message for i in range(len(warn))]
    warning_2 = warning_list[0]
    assert warning_2.args[0] == (
        "Incomplete Request. Index could not untrack baskets, "
        "as some were not being tracked to begin with."
    )
    assert warning_2.args[1] == 1


def test_index_abc_untrack_basket_removes_multiple_baskets(test_pantry):
    """Test IndexABC untrack_basket works when passing a multi-address list."""
    # Unpack the test_pantry into two variables for the pantry and index.
    test_pantry, ind = test_pantry

    # Put basket in the temporary pantry.
    tmp_basket_dir_one = test_pantry.set_up_basket("basket_one")
    test_pantry.upload_basket(tmp_basket_dir=tmp_basket_dir_one, uid="0001")

    tmp_basket_dir_one = test_pantry.set_up_basket("basket_two")
    test_pantry.upload_basket(tmp_basket_dir=tmp_basket_dir_one, uid="0002")

    tmp_basket_dir_one = test_pantry.set_up_basket("basket_three")
    test_pantry.upload_basket(tmp_basket_dir=tmp_basket_dir_one, uid="0003")

    # Generate the index.
    ind.generate_index()

    # Create a list of addresses to remove the first two baskets.
    untrack_uuids = ["0001", "0002"]

    ind.untrack_basket(untrack_uuids)
    assert len(ind) == 1, "untrack_basket([uuid]) failed to update index."


def test_index_abc_get_rows_single_address_works(test_pantry):
    """Test IndexABC get_rows returns manifest data of a single address."""
    # Unpack the test_pantry into two variables for the pantry and index.
    test_pantry, ind = test_pantry

    basket_type = "test_basket_type"

    uuids = ["0001"]
    labels = ["label_1"]
    parent_ids = [[]]

    # Put basket in the temporary pantry.
    tmp_basket_dir_one = test_pantry.set_up_basket("basket_one")
    up_dir1 = test_pantry.upload_basket(
        tmp_basket_dir=tmp_basket_dir_one,
        basket_type=basket_type,
        uid=uuids[-1],
        label=labels[-1],
        parent_ids=parent_ids[-1],
    )

    uuids.append("0002")
    labels.append("label_2")
    parent_ids.append([uuids[0]])

    tmp_basket_dir_two = test_pantry.set_up_basket("basket_two")
    up_dir2 = test_pantry.upload_basket(
        tmp_basket_dir=tmp_basket_dir_two,
        basket_type=basket_type,
        uid=uuids[-1],
        label=labels[-1],
        parent_ids=parent_ids[-1],
    )

    tmp_basket_dir_three = test_pantry.set_up_basket("basket_three")
    test_pantry.upload_basket(tmp_basket_dir=tmp_basket_dir_three, uid="0003")

    # Generate the index.
    ind.generate_index()

    first_row_df = ind.get_rows("0001")
    assert isinstance(first_row_df, pd.DataFrame) and len(first_row_df) == 1
    assert (
        first_row_df.iloc[0]["uuid"] == uuids[0] and
        isinstance(first_row_df.iloc[0]["upload_time"], datetime) and
        first_row_df.iloc[0]["parent_uuids"] == parent_ids[0] and
        first_row_df.iloc[0]["basket_type"] == basket_type and
        first_row_df.iloc[0]["label"] == labels[0] and
        first_row_df.iloc[0]["address"].endswith(up_dir1) and
        first_row_df.iloc[0]["storage_type"] == \
            test_pantry.file_system.__class__.__name__
    ), "Retrieved manifest values do not match first record."

    second_row_df = ind.get_rows(up_dir2)
    assert isinstance(second_row_df, pd.DataFrame) and len(second_row_df) == 1
    assert (
        second_row_df.iloc[0]["uuid"] == uuids[1] and
        isinstance(second_row_df.iloc[0]["upload_time"], datetime) and
        second_row_df.iloc[0]["parent_uuids"] == parent_ids[1] and
        second_row_df.iloc[0]["basket_type"] == basket_type and
        second_row_df.iloc[0]["label"] == labels[1] and
        second_row_df.iloc[0]["address"].endswith(up_dir2) and
        second_row_df.iloc[0]["storage_type"] == \
            test_pantry.file_system.__class__.__name__
    ), "Retrieved manifest values do not match second record."


def test_index_abc_get_rows_multiple_address_works(test_pantry):
    """Test IndexABC get_rows returns manifest data of multiple addresses."""
    # Unpack the test_pantry into two variables for the pantry and index.
    test_pantry, ind = test_pantry

    basket_type = "test_basket_type"

    uuid1 = "0001"
    label1 = "label_1"
    parent_ids1 = []

    # Put basket in the temporary pantry.
    tmp_basket_dir_one = test_pantry.set_up_basket("basket_one")
    up_dir1 = test_pantry.upload_basket(
        tmp_basket_dir=tmp_basket_dir_one,
        basket_type=basket_type,
        uid=uuid1,
        label=label1,
        parent_ids=parent_ids1,
    )

    uuid2 = "0002"
    label2 = "label_2"
    parent_ids2 = [uuid1]

    tmp_basket_dir_two = test_pantry.set_up_basket("basket_two")
    up_dir2 = test_pantry.upload_basket(
        tmp_basket_dir=tmp_basket_dir_two,
        basket_type=basket_type,
        uid=uuid2,
        label=label2,
        parent_ids=parent_ids2,
    )

    tmp_basket_dir_three = test_pantry.set_up_basket("basket_three")
    test_pantry.upload_basket(tmp_basket_dir=tmp_basket_dir_three, uid="0003")

    # Generate the index.
    ind.generate_index()

    rows_df = ind.get_rows(["0001", "0002"])
    assert isinstance(rows_df, pd.DataFrame) and len(rows_df) == 2
    assert (
        rows_df.iloc[0]["uuid"] == uuid1 and
        isinstance(rows_df.iloc[0]["upload_time"], datetime) and
        rows_df.iloc[0]["parent_uuids"] == parent_ids1 and
        rows_df.iloc[0]["basket_type"] == basket_type and
        rows_df.iloc[0]["label"] == label1 and
        rows_df.iloc[0]["address"].endswith(up_dir1) and
        rows_df.iloc[0]["storage_type"] == \
            test_pantry.file_system.__class__.__name__
    ), "Retrieved manifest values do not match first record."
    assert (
        rows_df.iloc[1]["uuid"] == uuid2 and
        isinstance(rows_df.iloc[1]["upload_time"], datetime) and
        rows_df.iloc[1]["parent_uuids"] == parent_ids2 and
        rows_df.iloc[1]["basket_type"] == basket_type and
        rows_df.iloc[1]["label"] == label2 and
        rows_df.iloc[1]["address"].endswith(up_dir2) and
        rows_df.iloc[1]["storage_type"] == \
            test_pantry.file_system.__class__.__name__
    ), "Retrieved manifest values do not match second record."


def test_index_abc_get_parents_path_works(test_pantry):
    """Test IndexABC get_parents(path) returns proper structure and values."""
    # Unpack the test_pantry into two variables for the pantry and index.
    test_pantry, ind = test_pantry

    # Setup random strucutre of parents and children
    tmp_dir = test_pantry.set_up_basket("great_grandparent_3")
    test_pantry.upload_basket(tmp_basket_dir=tmp_dir, uid="3000")

    tmp_dir = test_pantry.set_up_basket("great_grandparent_3_1")
    test_pantry.upload_basket(tmp_basket_dir=tmp_dir, uid="3003")

    tmp_dir = test_pantry.set_up_basket("great_grandparent_3_2")
    test_pantry.upload_basket(tmp_basket_dir=tmp_dir, uid="3333")

    tmp_dir = test_pantry.set_up_basket("great_grandparent_3_3")
    test_pantry.upload_basket(tmp_basket_dir=tmp_dir, uid="3303")

    tmp_dir = test_pantry.set_up_basket("grandparent_2")
    test_pantry.upload_basket(
        tmp_basket_dir=tmp_dir, uid="2000", parent_ids=["3000", "3003", "3333"]
    )

    tmp_dir = test_pantry.set_up_basket("grandparent_2_1")
    test_pantry.upload_basket(tmp_basket_dir=tmp_dir, uid="2002")

    tmp_dir = test_pantry.set_up_basket("parent_1")
    test_pantry.upload_basket(
        tmp_basket_dir=tmp_dir, uid="1000", parent_ids=["2000", "2002", "3303"]
    )

    tmp_dir = test_pantry.set_up_basket("parent_1_1")
    test_pantry.upload_basket(tmp_basket_dir=tmp_dir, uid="1001")

    tmp_dir = test_pantry.set_up_basket("child_0")
    child = test_pantry.upload_basket(
        tmp_basket_dir=tmp_dir, uid="0000", parent_ids=["1001", "1000"]
    )

    # String to shorten things for ruff
    gen_lvl = "generation_level"

    ind.generate_index()

    # Setup df of the right answer
    parent_ids = [
        "1000",
        "1001",
        "2000",
        "2002",
        "3303",
        "3000",
        "3003",
        "3333",
    ]
    parent_gens = [1, 1, 2, 2, 2, 3, 3, 3]
    index = ind.to_pandas_df()
    parent_answer = index.loc[index["uuid"].isin(parent_ids)]

    # Pandas wants to make a copy before adding a column
    # Used to remove warning in pytest
    parent_answer = parent_answer.copy()
    # Add the generation levels
    for i, j in zip(parent_ids, parent_gens):
        parent_answer.loc[parent_answer["uuid"] == i, gen_lvl] = j

    # Get the results
    results = ind.get_parents(child)

    # Sort so that they can be properly compared to
    parent_answer = parent_answer.sort_values(by="uuid").reset_index(drop=True)
    results = results.sort_values(by="uuid").reset_index(drop=True)

    # Cast to int64 so datatypes match
    parent_answer[gen_lvl] = parent_answer[gen_lvl].astype(np.int64)

    assert parent_answer.equals(results)


def test_index_abc_get_parents_uuid_works(test_pantry):
    """Test IndexABC get_parents(uuid) returns proper structure and values."""
    # Unpack the test_pantry into two variables for the pantry and index.
    test_pantry, ind = test_pantry

    # Setup random strucutre of parents and children
    tmp_dir = test_pantry.set_up_basket("great_grandparent_3")
    test_pantry.upload_basket(tmp_basket_dir=tmp_dir, uid="3000")

    tmp_dir = test_pantry.set_up_basket("great_grandparent_3_1")
    test_pantry.upload_basket(tmp_basket_dir=tmp_dir, uid="3003")

    tmp_dir = test_pantry.set_up_basket("great_grandparent_3_2")
    test_pantry.upload_basket(tmp_basket_dir=tmp_dir, uid="3333")

    tmp_dir = test_pantry.set_up_basket("great_grandparent_3_3")
    test_pantry.upload_basket(tmp_basket_dir=tmp_dir, uid="3303")

    tmp_dir = test_pantry.set_up_basket("grandparent_2")
    test_pantry.upload_basket(
        tmp_basket_dir=tmp_dir, uid="2000", parent_ids=["3000", "3003", "3333"]
    )

    tmp_dir = test_pantry.set_up_basket("grandparent_2_1")
    test_pantry.upload_basket(tmp_basket_dir=tmp_dir, uid="2002")

    tmp_dir = test_pantry.set_up_basket("parent_1")
    test_pantry.upload_basket(
        tmp_basket_dir=tmp_dir, uid="1000", parent_ids=["2000", "2002", "3303"]
    )

    tmp_dir = test_pantry.set_up_basket("parent_1_1")
    test_pantry.upload_basket(tmp_basket_dir=tmp_dir, uid="1001")

    tmp_dir = test_pantry.set_up_basket("child_0")
    test_pantry.upload_basket(
        tmp_basket_dir=tmp_dir, uid="0000", parent_ids=["1001", "1000"]
    )

    # String to shorten things for ruff
    gen_lvl = "generation_level"

    ind.generate_index()

    # Setup df of the right answer
    parent_ids = [
        "1000",
        "1001",
        "2000",
        "2002",
        "3303",
        "3000",
        "3003",
        "3333",
    ]
    parent_gens = [1, 1, 2, 2, 2, 3, 3, 3]
    index = ind.to_pandas_df()
    parent_answer = index.loc[index["uuid"].isin(parent_ids)]

    # Pandas wants to make a copy before adding a column
    # Used to remove warning in pytest
    parent_answer = parent_answer.copy()
    # Add the generation levels
    for i, j in zip(parent_ids, parent_gens):
        parent_answer.loc[parent_answer["uuid"] == i, gen_lvl] = j

    # Get the results
    results = ind.get_parents("0000")

    # Sort so that they can be properly compared to
    parent_answer = parent_answer.sort_values(by="uuid").reset_index(drop=True)
    results = results.sort_values(by="uuid").reset_index(drop=True)

    # Cast to int64 so datatypes match
    parent_answer[gen_lvl] = parent_answer[gen_lvl].astype(np.int64)

    assert parent_answer.equals(results)


def test_index_abc_get_parents_invalid_basket_address(test_pantry):
    """Test IndexABC get_parents fails given an invalid basket address."""
    # Unpack the test_pantry into two variables for the pantry and index.
    test_pantry, ind = test_pantry

    basket_path = "INVALIDpath"

    with pytest.raises(
        FileNotFoundError,
        match=f"basket path or uuid does not exist '{basket_path}'",
    ):
        ind.get_parents(basket_path)


def test_index_abc_get_parents_no_parents(test_pantry):
    """Test IndexABC get_parents returns an empty dataframe when a basket has
    no parents.
    """
    # Unpack the test_pantry into two variables for the pantry and index.
    test_pantry, ind = test_pantry

    no_parents = test_pantry.set_up_basket("no_parents")
    no_parents_path = test_pantry.upload_basket(
        tmp_basket_dir=no_parents, uid="0001"
    )

    ind.generate_index()

    parent_indeces = ind.get_parents(no_parents_path)

    assert parent_indeces.empty


def test_index_abc_get_parents_parent_is_child_loop(test_pantry):
    """Test IndexABC get_parents fails if parent-child loop exists.

    Set up 3 baskets, child, parent, grandparent, but the grandparent's
    parent_ids has the child's uid. This causes an infinite loop,
    check that it throws an error.
    """

    # Unpack the test_pantry into two variables for the pantry and index.
    test_pantry, ind = test_pantry

    # Create a basket structure with child, parent, and grandparent, but
    # the grandparent's parent, is the child, making an loop for the
    # parent-child relationship
    tmp_dir = test_pantry.set_up_basket("grandparent")
    test_pantry.upload_basket(
        tmp_basket_dir=tmp_dir, uid="3000", parent_ids=["1000"]
    )

    tmp_dir = test_pantry.set_up_basket("parent")
    test_pantry.upload_basket(
        tmp_basket_dir=tmp_dir, uid="2000", parent_ids=["3000"]
    )

    tmp_dir = test_pantry.set_up_basket("child")
    child = test_pantry.upload_basket(
        tmp_basket_dir=tmp_dir, uid="1000", parent_ids=["2000"]
    )

    ind.generate_index()

    fail = "1000"

    with pytest.raises(
        ValueError, match=f"Parent-Child loop found at uuid: {fail}"
    ):
        ind.get_parents(child)


def test_index_abc_get_parents_15_deep(test_pantry):
    """Test IndexABC get_parents works with deep parent-child structures.

    Make a parent-child relationship of baskets 15 deep, get all the parents
    Pass a child with a great*15 grandparent, and return all the grandparents
    for the child.

    Manually make the data and compare with the result.
    """
    # Unpack the test_pantry into two variables for the pantry and index.
    test_pantry, ind = test_pantry

    parent_id = "x"

    for i in range(15):
        child_id = parent_id
        parent_id = str(i)
        tmp = test_pantry.set_up_basket("basket_" + child_id)
        test_pantry.upload_basket(
            tmp_basket_dir=tmp, uid=child_id, parent_ids=[parent_id]
        )

    ind.generate_index()
    index = ind.to_pandas_df()

    child_path = index.loc[index["uuid"] == "x"]["address"].values[0]

    results = ind.get_parents(child_path)

    # Get the anwser to compare to the results
    par_ids = [
        "0",
        "1",
        "2",
        "3",
        "4",
        "5",
        "6",
        "7",
        "8",
        "9",
        "10",
        "11",
        "12",
        "13",
    ]
    par_gens = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14]
    answer = index.loc[index["uuid"].isin(par_ids)]

    gen_lvl = "generation_level"

    # Pandas wants to make a copy before adding a column
    # Used to remove warning in pytest
    answer = answer.copy()
    for i, j in zip(par_ids, par_gens):
        answer.loc[answer["uuid"] == i, gen_lvl] = j

    # Format and sort so .equals can be properly used
    answer = answer.sort_values(by="uuid").reset_index(drop=True)
    results = results.sort_values(by="uuid").reset_index(drop=True)
    answer[gen_lvl] = answer[gen_lvl].astype(np.int64)
    assert answer.equals(results)


def test_index_abc_get_parents_complex_fail(test_pantry):
    """Test IndexABC get_parents fails on an invalid complicated loop tree."""
    # Unpack the test_pantry into two variables for the pantry and index.
    test_pantry, ind = test_pantry

    tmp_dir = test_pantry.set_up_basket("parent_8")
    test_pantry.upload_basket(
        tmp_basket_dir=tmp_dir, uid="008", parent_ids=["007"]
    )

    tmp_dir = test_pantry.set_up_basket("parent_7")
    test_pantry.upload_basket(
        tmp_basket_dir=tmp_dir, uid="007", parent_ids=["000"]
    )

    tmp_dir = test_pantry.set_up_basket("parent_6")
    test_pantry.upload_basket(
        tmp_basket_dir=tmp_dir, uid="006", parent_ids=["008"]
    )

    tmp_dir = test_pantry.set_up_basket("parent_5")
    test_pantry.upload_basket(
        tmp_basket_dir=tmp_dir, uid="005", parent_ids=["007"]
    )

    tmp_dir = test_pantry.set_up_basket("parent_4")
    test_pantry.upload_basket(tmp_basket_dir=tmp_dir, uid="004")

    tmp_dir = test_pantry.set_up_basket("parent_3")
    test_pantry.upload_basket(
        tmp_basket_dir=tmp_dir, uid="003", parent_ids=["006"]
    )

    tmp_dir = test_pantry.set_up_basket("parent_2")
    test_pantry.upload_basket(
        tmp_basket_dir=tmp_dir, uid="002", parent_ids=["0004", "005", "008"]
    )

    tmp_dir = test_pantry.set_up_basket("parent_1")
    test_pantry.upload_basket(
        tmp_basket_dir=tmp_dir, uid="001", parent_ids=["004"]
    )

    tmp_dir = test_pantry.set_up_basket("child")
    child_path = test_pantry.upload_basket(
        tmp_basket_dir=tmp_dir, uid="000", parent_ids=["001", "002", "003"]
    )

    ind.generate_index()

    with pytest.raises(
        ValueError, match=re.escape("Parent-Child loop found at uuid: 000")
    ):
        ind.get_parents(child_path)


def test_index_abc_get_parents_max_gen_level(test_pantry):
    """Test IndexABC get_parents max_gen_level removes uuids where the
    gen_level is greater than the max_gen_level.
    """

    # Unpack the test_pantry into two variables for the pantry and index.
    test_pantry, ind = test_pantry

    # Setup random strucutre of parents and children
    tmp_dir = test_pantry.set_up_basket("great_grandparent_3")
    test_pantry.upload_basket(tmp_basket_dir=tmp_dir, uid="3000")

    tmp_dir = test_pantry.set_up_basket("great_grandparent_3_1")
    test_pantry.upload_basket(tmp_basket_dir=tmp_dir, uid="3003")

    tmp_dir = test_pantry.set_up_basket("great_grandparent_3_2")
    test_pantry.upload_basket(tmp_basket_dir=tmp_dir, uid="3333")

    tmp_dir = test_pantry.set_up_basket("great_grandparent_3_3")
    test_pantry.upload_basket(tmp_basket_dir=tmp_dir, uid="3303")

    tmp_dir = test_pantry.set_up_basket("grandparent_2")
    test_pantry.upload_basket(
        tmp_basket_dir=tmp_dir, uid="2000", parent_ids=["3000", "3003", "3333"]
    )

    tmp_dir = test_pantry.set_up_basket("grandparent_2_1")
    test_pantry.upload_basket(tmp_basket_dir=tmp_dir, uid="2002")

    tmp_dir = test_pantry.set_up_basket("parent_1")
    test_pantry.upload_basket(
        tmp_basket_dir=tmp_dir, uid="1000", parent_ids=["2000", "2002", "3303"]
    )

    tmp_dir = test_pantry.set_up_basket("parent_1_1")
    test_pantry.upload_basket(tmp_basket_dir=tmp_dir, uid="1001")

    tmp_dir = test_pantry.set_up_basket("child_0")
    test_pantry.upload_basket(
        tmp_basket_dir=tmp_dir, uid="0000", parent_ids=["1001", "1000"]
    )

    ind.generate_index()

    parents_df = ind.get_parents("0000", max_gen_level=2)
    parents_df_uuids = parents_df["uuid"].tolist()
    parents_df_uuids.sort()

    correct_uuids = ["1000", "1001", "2000", "2002", "3303"]
    correct_uuids.sort()

    # Check that max_gen_level filters out the correct uuids
    assert parents_df_uuids == correct_uuids


def test_index_abc_get_children_path_works(test_pantry):
    """Test IndexABC get_children(path) returns proper structure and values."""
    # Unpack the test_pantry into two variables for the pantry and index.
    test_pantry, ind = test_pantry

    # Setup random strucutre of parents and children
    tmp_dir = test_pantry.set_up_basket("great_grandparent_3")
    great_grandparent = test_pantry.upload_basket(
        tmp_basket_dir=tmp_dir, uid="3000"
    )

    tmp_dir = test_pantry.set_up_basket("great_grandparent_3_1")
    test_pantry.upload_basket(tmp_basket_dir=tmp_dir, uid="3003")

    tmp_dir = test_pantry.set_up_basket("great_grandparent_3_2")
    test_pantry.upload_basket(tmp_basket_dir=tmp_dir, uid="3333")

    tmp_dir = test_pantry.set_up_basket("great_grandparent_3_3")
    test_pantry.upload_basket(tmp_basket_dir=tmp_dir, uid="3303")

    tmp_dir = test_pantry.set_up_basket("grandparent_2")
    test_pantry.upload_basket(
        tmp_basket_dir=tmp_dir, uid="2000", parent_ids=["3000", "3003", "3333"]
    )

    tmp_dir = test_pantry.set_up_basket("grandparent_2_1")
    test_pantry.upload_basket(tmp_basket_dir=tmp_dir, uid="2002")

    tmp_dir = test_pantry.set_up_basket("parent_1")
    test_pantry.upload_basket(
        tmp_basket_dir=tmp_dir, uid="1000", parent_ids=["2000", "2002", "3303"]
    )

    tmp_dir = test_pantry.set_up_basket("parent_1_1")
    test_pantry.upload_basket(tmp_basket_dir=tmp_dir, uid="1001")

    tmp_dir = test_pantry.set_up_basket("child_0")
    test_pantry.upload_basket(
        tmp_basket_dir=tmp_dir, uid="0000", parent_ids=["1001", "1000"]
    )

    # String to shorten things for ruff
    gen_lvl = "generation_level"

    ind.generate_index()

    # Setup df of the right answer
    child_ids = ["2000", "1000", "0000"]
    child_gens = [-1, -2, -3]
    index = ind.to_pandas_df()
    child_answer = index.loc[index["uuid"].isin(child_ids)]

    # Pandas wants to make a copy before adding a column
    # Used to remove warning in pytest
    child_answer = child_answer.copy()
    # Add the generation levels
    for i, j in zip(child_ids, child_gens):
        child_answer.loc[child_answer["uuid"] == i, gen_lvl] = j

    # Get the results
    results = ind.get_children(great_grandparent)

    # Sort so that they can be properly compared to
    child_answer = child_answer.sort_values(by="uuid").reset_index(drop=True)
    results = results.sort_values(by="uuid").reset_index(drop=True)

    # Cast to int64 so datatypes match
    child_answer[gen_lvl] = child_answer[gen_lvl].astype(np.int64)
    assert child_answer.equals(results)


def test_index_abc_get_children_uuid_works(test_pantry):
    """Test IndexABC get_children(uuid) returns proper structure and values."""
    # Unpack the test_pantry into two variables for the pantry and index.
    test_pantry, ind = test_pantry

    # Setup random strucutre of parents and children
    tmp_dir = test_pantry.set_up_basket("great_grandparent_3")
    test_pantry.upload_basket(
        tmp_basket_dir=tmp_dir, uid="3000"
    )

    tmp_dir = test_pantry.set_up_basket("great_grandparent_3_1")
    test_pantry.upload_basket(tmp_basket_dir=tmp_dir, uid="3003")

    tmp_dir = test_pantry.set_up_basket("great_grandparent_3_2")
    test_pantry.upload_basket(tmp_basket_dir=tmp_dir, uid="3333")

    tmp_dir = test_pantry.set_up_basket("great_grandparent_3_3")
    test_pantry.upload_basket(tmp_basket_dir=tmp_dir, uid="3303")

    tmp_dir = test_pantry.set_up_basket("grandparent_2")
    test_pantry.upload_basket(
        tmp_basket_dir=tmp_dir, uid="2000", parent_ids=["3000", "3003", "3333"]
    )

    tmp_dir = test_pantry.set_up_basket("grandparent_2_1")
    test_pantry.upload_basket(tmp_basket_dir=tmp_dir, uid="2002")

    tmp_dir = test_pantry.set_up_basket("parent_1")
    test_pantry.upload_basket(
        tmp_basket_dir=tmp_dir, uid="1000", parent_ids=["2000", "2002", "3303"]
    )

    tmp_dir = test_pantry.set_up_basket("parent_1_1")
    test_pantry.upload_basket(tmp_basket_dir=tmp_dir, uid="1001")

    tmp_dir = test_pantry.set_up_basket("child_0")
    test_pantry.upload_basket(
        tmp_basket_dir=tmp_dir, uid="0000", parent_ids=["1001", "1000"]
    )

    # String to shorten things for ruff
    gen_lvl = "generation_level"

    ind.generate_index()

    # Setup df of the right answer
    child_ids = ["2000", "1000", "0000"]
    child_gens = [-1, -2, -3]
    index = ind.to_pandas_df()
    child_answer = index.loc[index["uuid"].isin(child_ids)]

    # Pandas wants to make a copy before adding a column
    # Used to remove warning in pytest
    child_answer = child_answer.copy()
    # Add the generation levels
    for i, j in zip(child_ids, child_gens):
        child_answer.loc[child_answer["uuid"] == i, gen_lvl] = int(j)

    # Get the results
    results = ind.get_children("3000")

    # Sort so that they can be properly compared to
    child_answer = child_answer.sort_values(by="uuid").reset_index(drop=True)
    results = results.sort_values(by="uuid").reset_index(drop=True)

    # Cast to int64 so datatypes match
    child_answer[gen_lvl] = child_answer[gen_lvl].astype(np.int64)

    assert child_answer.equals(results)


def test_index_abc_get_children_invalid_basket_address(test_pantry):
    """Test IndexABC get_children fails given an invalid basket address."""
    test_pantry, ind = test_pantry

    basket_path = "INVALIDpath"

    with pytest.raises(
        FileNotFoundError,
        match=f"basket path or uuid does not exist '{basket_path}'",
    ):
        ind.get_children(basket_path)


def test_index_abc_get_children_no_children(test_pantry):
    """Test IndexABC get_children returns an empty dataframe when a basket has
    no children."""
    # Unpack the test_pantry into two variables for the pantry and index.
    test_pantry, ind = test_pantry

    no_children = test_pantry.set_up_basket("no_children")
    no_children_path = test_pantry.upload_basket(
        tmp_basket_dir=no_children, uid="0001"
    )

    ind.generate_index()

    children_indexes = ind.get_children(no_children_path)

    assert children_indexes.empty


def test_index_abc_get_children_child_is_parent_loop(test_pantry):
    """Test IndexABC get_children fails if parent-child loop exists.

    Set up 3 baskets, child, parent, grandparent, but the grandparents's
    parent_ids has the child's uid. This causes an infinite loop,
    check that it throws an error.
    """
    # Unpack the test_pantry into two variables for the pantry and index.
    test_pantry, ind = test_pantry

    # Create a basket structure with child, parent, and grandparent, but
    # the grandparent's parent, is the child, making an loop for the
    # parent-child relationship
    tmp_dir = test_pantry.set_up_basket("grandparent")
    grandparent_basket = test_pantry.upload_basket(
        tmp_basket_dir=tmp_dir, uid="3000", parent_ids=["1000"]
    )

    tmp_dir = test_pantry.set_up_basket("parent")
    test_pantry.upload_basket(
        tmp_basket_dir=tmp_dir, uid="2000", parent_ids=["3000"]
    )

    tmp_dir = test_pantry.set_up_basket("child")
    test_pantry.upload_basket(
        tmp_basket_dir=tmp_dir, uid="1000", parent_ids=["2000"]
    )

    ind.generate_index()

    fail = "3000"

    with pytest.raises(
        ValueError, match=re.escape(f"Parent-Child loop found at uuid: {fail}")
    ):
        ind.get_children(grandparent_basket)


def test_index_abc_get_children_15_deep(test_pantry):
    """Test IndexABC get_children works with deep parent-child structures.

    Make a parent-child relationship of baskets 15 deep, get the children.
    Pass a child with great*15 grandparent, and return all the grandchildren
    for the highest grandparent.
    Manually make the data and compare with the result.
    """
    # Unpack the test_pantry into two variables for the pantry and index.
    test_pantry, ind = test_pantry
    parent_id = "x"

    for i in range(15):
        child_id = parent_id
        parent_id = f"00{i}"
        tmp = test_pantry.set_up_basket("basket_" + child_id)
        test_pantry.upload_basket(
            tmp_basket_dir=tmp, uid=child_id, parent_ids=[parent_id]
        )

    ind.generate_index()
    index = ind.to_pandas_df()
    results = ind.get_children("0013")

    # Get the anwser to compare to the results
    child_ids = [
        "x",
        "000",
        "001",
        "002",
        "003",
        "004",
        "005",
        "006",
        "007",
        "008",
        "009",
        "0010",
        "0011",
        "0012",
    ]
    child_gens = [-14, -13, -12, -11, -10, -9, -8, -7, -6, -5, -4, -3, -2, -1]
    answer = index.loc[index["uuid"].isin(child_ids)]

    gen_lvl = "generation_level"

    # Pandas wants to make a copy before adding a column
    # Used to remove warning in pytest
    answer = answer.copy()
    for i, j in zip(child_ids, child_gens):
        answer.loc[answer["uuid"] == i, gen_lvl] = j

    # Format and sort so .equals can be properly used
    answer = answer.sort_values(by="uuid").reset_index(drop=True)
    results = results.sort_values(by="uuid").reset_index(drop=True)
    answer[gen_lvl] = answer[gen_lvl].astype(np.int64)
    assert answer.equals(results)


def test_index_abc_get_children_complex_fail(test_pantry):
    """Test IndexABC get_children fails on an invalid complicated loop tree.
    Make a complicated tree with a loop to test new algorithm.
    """
    # Unpack the test_pantry into two variables for the pantry and index.
    test_pantry, ind = test_pantry

    tmp_dir = test_pantry.set_up_basket("parent_8")
    test_pantry.upload_basket(
        tmp_basket_dir=tmp_dir, uid="008", parent_ids=["007"]
    )

    tmp_dir = test_pantry.set_up_basket("parent_7")
    parent_path = test_pantry.upload_basket(
        tmp_basket_dir=tmp_dir, uid="007", parent_ids=["003"]
    )

    tmp_dir = test_pantry.set_up_basket("parent_6")
    test_pantry.upload_basket(
        tmp_basket_dir=tmp_dir, uid="006", parent_ids=["008"]
    )

    tmp_dir = test_pantry.set_up_basket("parent_5")
    test_pantry.upload_basket(
        tmp_basket_dir=tmp_dir, uid="005", parent_ids=["007"]
    )

    tmp_dir = test_pantry.set_up_basket("parent_4")
    test_pantry.upload_basket(tmp_basket_dir=tmp_dir, uid="004")

    tmp_dir = test_pantry.set_up_basket("parent_3")
    test_pantry.upload_basket(
        tmp_basket_dir=tmp_dir, uid="003", parent_ids=["006"]
    )

    tmp_dir = test_pantry.set_up_basket("parent_2")
    test_pantry.upload_basket(
        tmp_basket_dir=tmp_dir, uid="002", parent_ids=["004", "005", "008"]
    )

    tmp_dir = test_pantry.set_up_basket("parent_1")
    test_pantry.upload_basket(
        tmp_basket_dir=tmp_dir, uid="001", parent_ids=["004"]
    )

    tmp_dir = test_pantry.set_up_basket("child")
    test_pantry.upload_basket(
        tmp_basket_dir=tmp_dir, uid="000", parent_ids=["001", "002", "003"]
    )

    ind.generate_index()

    with pytest.raises(
        ValueError, match=re.escape("Parent-Child loop found at uuid: 007")
    ):
        ind.get_children(parent_path)


def test_index_abc_get_children_min_gen_level(test_pantry):
    """Test IndexABC get_children min_gen_level removes uuids where the
    gen_level is less than the min_gen_level.
    """

    # Unpack the test_pantry into two variables for the pantry and index.
    test_pantry, ind = test_pantry

    # Setup random strucutre of parents and children
    tmp_dir = test_pantry.set_up_basket("great_grandparent_3")
    test_pantry.upload_basket(tmp_basket_dir=tmp_dir, uid="3000")

    tmp_dir = test_pantry.set_up_basket("great_grandparent_3_1")
    test_pantry.upload_basket(tmp_basket_dir=tmp_dir, uid="3003")

    tmp_dir = test_pantry.set_up_basket("great_grandparent_3_2")
    test_pantry.upload_basket(tmp_basket_dir=tmp_dir, uid="3333")

    tmp_dir = test_pantry.set_up_basket("great_grandparent_3_3")
    test_pantry.upload_basket(tmp_basket_dir=tmp_dir, uid="3303")

    tmp_dir = test_pantry.set_up_basket("grandparent_2")
    test_pantry.upload_basket(
        tmp_basket_dir=tmp_dir, uid="2000", parent_ids=["3000", "3003", "3333"]
    )

    tmp_dir = test_pantry.set_up_basket("grandparent_2_1")
    test_pantry.upload_basket(tmp_basket_dir=tmp_dir, uid="2002")

    tmp_dir = test_pantry.set_up_basket("parent_1")
    test_pantry.upload_basket(
        tmp_basket_dir=tmp_dir, uid="1000", parent_ids=["2000", "2002", "3303"]
    )

    tmp_dir = test_pantry.set_up_basket("parent_1_1")
    test_pantry.upload_basket(tmp_basket_dir=tmp_dir, uid="1001")

    tmp_dir = test_pantry.set_up_basket("child_0")
    test_pantry.upload_basket(
        tmp_basket_dir=tmp_dir, uid="0000", parent_ids=["1001", "1000"]
    )

    ind.generate_index()

    children_df = ind.get_children("3000", min_gen_level=-2)
    children_df_uuids = children_df["uuid"].tolist()
    children_df_uuids.sort()

    correct_uuids = ["1000", "2000"]
    correct_uuids.sort()

    # Check that min_gen_level filters the correct uuids
    assert children_df_uuids == correct_uuids


def test_index_abc_get_baskets_of_type_works(test_pantry):
    """Test IndexABC get_baskets_of_type returns correct dataframe."""
    # Unpack the test_pantry into two variables for the pantry and index.
    test_pantry, ind = test_pantry

    # Put basket in the temporary pantry.
    tmp_basket_dir_one = test_pantry.set_up_basket("basket_one")
    test_pantry.upload_basket(
        tmp_basket_dir=tmp_basket_dir_one,
        uid="0001",
    )
    tmp_basket_dir_one = test_pantry.set_up_basket("basket_two")
    test_pantry.upload_basket(
        basket_type="other",
        tmp_basket_dir=tmp_basket_dir_one,
        uid="0002",
    )

    # Generate the index.
    ind.generate_index()

    baskets = ind.get_baskets_of_type("test_basket")
    assert len(baskets) == 1
    assert "0001" in baskets["uuid"].to_list()


def test_index_abc_get_baskets_of_type_max_rows_works(test_pantry):
    """Test IndexABC get_baskets_of_type max_rows argument works properly."""
    # Unpack the test_pantry into two variables for the pantry and index.
    test_pantry, ind = test_pantry

    # Put basket in the temporary pantry.
    for basket_iter in range(3):
        tmp_basket_dir_one = test_pantry.set_up_basket(f"basket_{basket_iter}")
        test_pantry.upload_basket(
            tmp_basket_dir=tmp_basket_dir_one,
            uid=f"000{basket_iter}",
        )

    # Generate the index.
    ind.generate_index()

    baskets = ind.get_baskets_of_type("test_basket", max_rows=2)
    assert len(baskets) == 2


def test_index_abc_get_baskets_of_type_returns_empty_df(test_pantry):
    """Test IndexABC get_baskets_of_type returns empty df if no baskets of
    type.
    """
    # Unpack the test_pantry into two variables for the pantry and index.
    test_pantry, ind = test_pantry

    # Put basket in the temporary pantry.
    tmp_basket_dir_one = test_pantry.set_up_basket("basket_one")
    test_pantry.upload_basket(
        tmp_basket_dir=tmp_basket_dir_one,
        uid="0001",
    )

    # Generate the index.
    ind.generate_index()

    baskets = ind.get_baskets_of_type("bad_basket_type")
    assert len(baskets) == 0


def test_index_abc_get_baskets_of_label_works(test_pantry):
    """Test IndexABC get_baskets_of_label returns correct dataframe."""
    # Unpack the test_pantry into two variables for the pantry and index.
    test_pantry, ind = test_pantry

    # Put basket in the temporary pantry.
    tmp_basket_dir_one = test_pantry.set_up_basket("basket_one")
    test_pantry.upload_basket(
        tmp_basket_dir=tmp_basket_dir_one,
        uid="0001",
        label='good_label',
    )
    tmp_basket_dir_one = test_pantry.set_up_basket("basket_two")
    test_pantry.upload_basket(
        tmp_basket_dir=tmp_basket_dir_one,
        uid="0002",
        label='bad_label',
    )

    # Generate the index.
    ind.generate_index()

    baskets = ind.get_baskets_of_label("good_label")
    assert len(baskets) == 1
    assert "0001" in baskets["uuid"].to_list()


def test_index_abc_get_baskets_of_label_max_rows_works(test_pantry):
    """Test IndexABC get_baskets_of_label max_rows argument works properly."""
    # Unpack the test_pantry into two variables for the pantry and index.
    test_pantry, ind = test_pantry

    # Put basket in the temporary pantry.
    for basket_iter in range(3):
        tmp_basket_dir_one = test_pantry.set_up_basket(f"basket_{basket_iter}")
        test_pantry.upload_basket(
            tmp_basket_dir=tmp_basket_dir_one,
            uid=f"000{basket_iter}",
            label='good_label',
        )

    # Generate the index.
    ind.generate_index()

    baskets = ind.get_baskets_of_label("good_label", max_rows=2)
    assert len(baskets) == 2


def test_index_abc_get_baskets_of_label_returns_empty_df(test_pantry):
    """Test IndexABC get_baskets_of_label returns empty df if no baskets have
    the given label.
    """
    # Unpack the test_pantry into two variables for the pantry and index.
    test_pantry, ind = test_pantry

    # Put basket in the temporary pantry.
    tmp_basket_dir_one = test_pantry.set_up_basket("basket_one")
    test_pantry.upload_basket(
        tmp_basket_dir=tmp_basket_dir_one,
        uid="0001",
        label='good_label',
    )

    # Generate the index.
    ind.generate_index()

    baskets = ind.get_baskets_of_label("bad_label")
    assert len(baskets) == 0


def test_index_abc_get_baskets_by_upload_time_raises_value_error2(test_pantry):
    """Test IndexABC get_baskets_by_upload_time raises a ValueError when
    either start or stop times are not valid datetime format (ie, not UTC).
    """
    # Unpack the test_pantry into two variables for the pantry and index.
    test_pantry, ind = test_pantry

    with pytest.raises(ValueError, match="start_time is not datetime object."):
        ind.get_baskets_by_upload_time(start_time=1)
    with pytest.raises(ValueError, match="end_time is not datetime object."):
        time_str = "2022-02-03 11:12:13"
        ind.get_baskets_by_upload_time(end_time=time_str)


def test_index_abc_get_baskets_by_upload_time_start_works(test_pantry):
    """Test IndexABC get_baskets_by_upload_time works with only a start time.
    """
    # Unpack the test_pantry into two variables for the pantry and index.
    test_pantry, ind = test_pantry

    # Save the current time, and set the 'end' time to 5 seconds ago.
    start = datetime.now(timezone.utc) - timedelta(seconds=5)

    # Get the index columns and instantiate with empty lists.
    columns = weave.config.get_index_column_names()
    manifest_dict = dict.fromkeys(columns, ["placeholder-data"])

    # Set placeholder values (for columns that matter)
    # that will have the time and uuid replaced.
    manifest_dict["uuid"] = ["PLACEHOLDER"]
    manifest_dict["upload_time"] = [start]
    manifest_dict["parent_uuids"] = [[]]
    manifest_dict["basket_type"] = ["test_basket_type"]
    manifest_dict["label"] = ["test_label"]
    manifest_dict["address"] = ["temp-pantry-address"]
    manifest_dict["storage_type"] = ["fake-storage"]

    # Create and track a record, with the upload time 1 second before start.
    manifest_dict["uuid"] = ["0001"]
    manifest_dict["upload_time"] = [start - timedelta(seconds=1)]
    basket_df = pd.DataFrame.from_dict(manifest_dict)
    ind.track_basket(basket_df)

    # Create and track a record, with the upload time at the start.
    manifest_dict["uuid"] = ["0002"]
    manifest_dict["upload_time"] = [start]
    basket_df = pd.DataFrame.from_dict(manifest_dict)
    ind.track_basket(basket_df)

    # Create and track a record, with the upload time 1 second after the start.
    manifest_dict["uuid"] = ["0003"]
    manifest_dict["upload_time"] = [start + timedelta(seconds=1)]
    basket_df = pd.DataFrame.from_dict(manifest_dict)
    ind.track_basket(basket_df)

    baskets = ind.get_baskets_by_upload_time(start_time=start)

    assert len(baskets) == 2
    assert "0001" not in baskets["uuid"].to_list()


def test_index_abc_get_baskets_by_upload_time_end_works(test_pantry):
    """Test IndexABC get_baskets_by_upload_time works with only an end time."""
    # Unpack the test_pantry into two variables for the pantry and index.
    test_pantry, ind = test_pantry
    ind.to_pandas_df()

    # Save the current time, and set the 'end' time to 5 seconds ago.
    end = datetime.now(timezone.utc) - timedelta(seconds=5)

    # Get the index columns and instantiate with empty lists.
    columns = weave.config.get_index_column_names()
    manifest_dict = dict.fromkeys(columns, ["placeholder-data"])

    # Set placeholder values (for columns that matter)
    # that will have the time and uuid replaced.
    manifest_dict["uuid"] = ["PLACEHOLDER"]
    manifest_dict["upload_time"] = [end]
    manifest_dict["parent_uuids"] = [[]]
    manifest_dict["basket_type"] = ["test_basket_type"]
    manifest_dict["label"] = ["test_label"]
    manifest_dict["address"] = ["temp-pantry-address"]
    manifest_dict["storage_type"] = ["fake-storage"]

    # Create and track a record, with the upload time 1 second before the end.
    manifest_dict["uuid"] = ["0001"]
    manifest_dict["upload_time"] = [end - timedelta(seconds=1)]
    basket_df = pd.DataFrame.from_dict(manifest_dict)
    ind.track_basket(basket_df)

    # Create and track a record, with the upload time at the end.
    manifest_dict["uuid"] = ["0002"]
    manifest_dict["upload_time"] = [end]
    basket_df = pd.DataFrame.from_dict(manifest_dict)
    ind.track_basket(basket_df)

    # Create and track a record, with the upload time 1 second after the end.
    manifest_dict["uuid"] = ["0003"]
    manifest_dict["upload_time"] = [end + timedelta(seconds=1)]
    basket_df = pd.DataFrame.from_dict(manifest_dict)
    ind.track_basket(basket_df)

    baskets = ind.get_baskets_by_upload_time(end_time=end)

    assert len(baskets) == 2
    assert "0003" not in baskets["uuid"].to_list()


def test_index_abc_get_baskets_by_upload_time_start_end_works(test_pantry):
    """Test IndexABC get_baskets_by_upload_time works with both start and end
    times supplied.
    """
    # Unpack the test_pantry into two variables for the pantry and index.
    test_pantry, ind = test_pantry
    ind.to_pandas_df()

    # Save the current time, and set the 'start' time to now.
    start = datetime.now(timezone.utc)

    # Get the index columns and instantiate with empty lists.
    columns = weave.config.get_index_column_names()
    manifest_dict = dict.fromkeys(columns, ["placeholder-data"])

    # Set placeholder values (for columns that matter)
    # that will have the time and uuid replaced.
    manifest_dict["uuid"] = ["PLACEHOLDER"]
    manifest_dict["upload_time"] = [start]
    manifest_dict["parent_uuids"] = [[]]
    manifest_dict["basket_type"] = ["test_basket_type"]
    manifest_dict["label"] = ["test_label"]
    manifest_dict["address"] = ["temp-pantry-address"]
    manifest_dict["storage_type"] = ["fake-storage"]

    # Create and track a record, with the upload time 1 second after the start.
    manifest_dict["uuid"] = ["0001"]
    manifest_dict["upload_time"] = [start]
    basket_df = pd.DataFrame.from_dict(manifest_dict)
    ind.track_basket(basket_df)

    # Create and track a record, with the upload time 1 second after the start.
    manifest_dict["uuid"] = ["0002"]
    manifest_dict["upload_time"] = [start + timedelta(seconds=1)]
    basket_df = pd.DataFrame.from_dict(manifest_dict)
    ind.track_basket(basket_df)

    # Make the stop time 2 seconds after the start.
    end = start + timedelta(seconds=2)

    # Create and track a record, with the upload time at the end time.
    manifest_dict["uuid"] = ["0003"]
    manifest_dict["upload_time"] = [end]
    basket_df = pd.DataFrame.from_dict(manifest_dict)
    ind.track_basket(basket_df)

    # Create and track a record, with the upload time 1 second after the end.
    manifest_dict["uuid"] = ["0004"]
    manifest_dict["upload_time"] = [end + timedelta(seconds=1)]
    basket_df = pd.DataFrame.from_dict(manifest_dict)
    ind.track_basket(basket_df)

    # Get the baskets inbetween start and end times (1 basket).
    baskets = ind.get_baskets_by_upload_time(start_time=start, end_time=end)

    assert len(baskets) == 3
    assert "0004" not in baskets["uuid"].to_list()


def test_index_abc_get_baskets_by_upload_time_returns_empty_df(test_pantry):
    """Test IndexABC get_baskets_by_upload_time returns empty df when no entry
    is found between start/end.
    """
    # Unpack the test_pantry into two variables for the pantry and index.
    test_pantry, ind = test_pantry

    # Save the current time, and set the 'start' time to now.
    # Save the end time to 2 seconds after the start.
    start = datetime.now(timezone.utc)
    end = start + timedelta(seconds=2)

    # Get the index columns and instantiate with empty lists.
    columns = weave.config.get_index_column_names()
    manifest_dict = dict.fromkeys(columns, ["placeholder-data"])

    # Set placeholder values (for columns that matter)
    # that will have the time and uuid replaced.
    manifest_dict["uuid"] = ["PLACEHOLDER"]
    manifest_dict["upload_time"] = [start]
    manifest_dict["parent_uuids"] = [[]]
    manifest_dict["basket_type"] = ["test_basket_type"]
    manifest_dict["label"] = ["test_label"]
    manifest_dict["address"] = ["temp-pantry-address"]
    manifest_dict["storage_type"] = ["fake-storage"]

    # Create and track a record, with the upload time 1 second before the start
    manifest_dict["uuid"] = ["0001"]
    manifest_dict["upload_time"] = [start - timedelta(seconds=1)]

    basket_df = pd.DataFrame.from_dict(manifest_dict)
    ind.track_basket(basket_df)

    # Create and track a record, with the upload time 1 second after the end.
    manifest_dict["uuid"] = ["0002"]
    manifest_dict["upload_time"] = [end + timedelta(seconds=1)]
    basket_df = pd.DataFrame.from_dict(manifest_dict)
    ind.track_basket(basket_df)

    # Get the baskets inbetween start and end times (1 basket).
    baskets = ind.get_baskets_by_upload_time(start_time=start, end_time=end)

    assert isinstance(baskets, pd.DataFrame) and len(baskets) == 0


def test_index_abc_columns_in_df_are_same_as_config_index_columns(test_pantry):
    """Test IndexABC tracks the same columns found in
    config.get_index_column_names().
    """
    # Unpack the test_pantry into two variables for the pantry and index.
    test_pantry, ind = test_pantry

    # Put basket in the temporary pantry.
    tmp_basket_dir_one = test_pantry.set_up_basket("basket_one")
    test_pantry.upload_basket(
        tmp_basket_dir=tmp_basket_dir_one,
        uid="0001",
        label='good_label',
    )

    ind.generate_index()
    ind_df = ind.to_pandas_df()
    ind_df_columns = list(ind_df.columns)

    # Get the columns in the schema, and add derived columns.
    index_columns = weave.config.get_index_column_names()

    # Sort both for comparisons.
    ind_df_columns.sort()
    index_columns.sort()

    assert ind_df_columns == index_columns
